# Author: Carsten Sachse 
# Copyright: EMBL (2010 - 2018)
# License: see license.txt for details
"""
Program to extract overlapping segments from micrographs 
"""
from EMAN2 import EMData
import os
import shutil
from spring.csinfrastr.csdatabase import SpringDataBase, base, CtfMicrographTable, HelixTable, SegmentTable, refine_base, \
    RefinementCycleSegmentTable, RefinementCycleHelixTable
from spring.csinfrastr.csfeatures import FeaturesSupport, Features
from spring.csinfrastr.cslogger import Logger
from spring.micprgs.scansplit import Micrograph
from spring.segment2d.segmentctfapply import SegmentCtfApplyPar, SegmentCtfApply
from spring.segment2d.segmentselect import SegmentSelect

from tabulate import tabulate

import numpy as np


class SegmentPar:
    """
    Class to initiate default dictionary with input parameters including help and range values and status dictionary
    """
    def __init__(self):
        # package/program identity
        self.package = 'emspring'
        self.progname = 'segment'
        self.proginfo = __doc__
        self.code_files = ['segment_prep', 'segment_int', 'segmentctfapply', self.progname, self.progname + '_mpi']

        self.segment_features = Features()
        self.feature_set = self.segment_features.setup(self)
        
        self.define_parameters_and_their_properties()
        self.define_program_states()
    

    def set_segment_or_window_features(self):
        self.feature_set = self.set_segment_coordinate_files(self.feature_set)
        self.feature_set = self.set_segment_size(self.feature_set)
        self.feature_set = self.segment_features.set_helix_width(self.feature_set)
        self.feature_set = self.set_segmentation_step(self.feature_set)
        self.feature_set = self.set_perturb_step_option(self.feature_set)


    def set_center_and_remove_ends_for_helices(self):
        self.feature_set = self.set_remove_helix_ends_option(self.feature_set)

    def set_helix_or_particle_centering(self):
        self.set_center_and_remove_ends_for_helices()
        self.feature_set = self.set_rotation_option(self.feature_set)
        self.feature_set = self.set_straightening_option(self.feature_set)

    def define_parameters_and_their_properties(self):
        self.feature_set = self.segment_features.set_inp_multiple_micrographs(self.feature_set)
        self.feature_set = self.segment_features.set_out_stack(self.feature_set)
    
        self.feature_set = self.segment_features.set_spring_db_option(self.feature_set)
        self.feature_set = self.segment_features.set_spring_path(self.feature_set)
        self.set_segment_or_window_features()
        self.feature_set = self.segment_features.set_pixelsize(self.feature_set)
        self.feature_set = self.set_ctf_correct(self.feature_set)
        self.feature_set = self.segment_features.set_binning_option(self.feature_set, image='segments')
        self.feature_set = self.segment_features.set_binning_factor(self.feature_set, image='segments')
        self.feature_set = self.set_invert_option(self.feature_set)
        self.feature_set = self.set_normalization_option(self.feature_set)
        self.feature_set = self.set_rows_normalization_option(self.feature_set)
        
        self.feature_set = self.segment_features.set_micrograph_and_helix_selection_criteria(self.feature_set)
        self.feature_set = self.segment_features.set_curvature_selection_criteria(self.feature_set)
        self.feature_set = self.segment_features.set_defocus_and_astigmatism_selection_criteria(self.feature_set)
        self.set_helix_or_particle_centering()

        self.feature_set = self.set_frame_processing_option(self.feature_set)
        self.feature_set = self.set_frame_first_and_last_frame(self.feature_set)
        self.feature_set = self.set_ref_db_frame_processing(self.feature_set)

        self.feature_set = self.segment_features.set_mpi(self.feature_set)
        self.feature_set = self.segment_features.set_ncpus(self.feature_set)
        self.feature_set = self.segment_features.set_temppath(self.feature_set)

        
    def define_program_states(self):
        self.feature_set.program_states['assign_reorganize']='Initialize micrographs and segments to convert them ' + \
        'into Spring\'s file structure'
        self.feature_set.program_states['single_out']='Single out individual helices from micrograph'
        self.feature_set.program_states['readmic']='Loading new micrograph'
        self.feature_set.program_states['center_segments']='Segments are centerd with respect to helix axis'
        self.feature_set.program_states['window_segment']='Windowing segments from micrograph'

    def set_segment_coordinate_files(self, feature_set):
        inp3 = 'Segment coordinates'
        feature_set.parameters[inp3] = 'scan034_boxes.txt'
        feature_set.properties[inp3] = feature_set.file_properties(1000, ['box', 'txt', 'star', 'db'], 'getFiles')
        feature_set.hints[inp3] = 'Input: file with identical name of corresponding micrograph (accepted file ' + \
        'formats EMAN\'s Helixboxer/Boxer, EMAN2\'s E2helixboxer and Bsoft filament parameters coordinates: {0}{1}). '.\
        format(os.extsep, FeaturesSupport().add_file_extensions_in_comma_separated_string(feature_set, inp3)) + \
        'When using the frame processing please specify a previously generated spring.db to provide the coordinates. ' + \
        'Make sure that helix paths are continuous. A helix path can follow a C- or S-path but must NOT form a U-turn.'\
        
        feature_set.level[inp3]='beginner'

        return feature_set


    def set_segment_size(self, feature_set):
        inp4 = 'Segment size in Angstrom'
        feature_set.parameters[inp4] = int(700)
        feature_set.hints[inp4] = 'Molecular mass (i.e. signal) increases with segment size and helix defects ' + \
        'become more pronounced. Final image size = segement size + stepsize.'
        feature_set.properties[inp4] = feature_set.Range(100, 1500, 50)
        feature_set.level[inp4]='beginner'
        
        return feature_set


    def set_segmentation_step(self, feature_set):
        inp5 = 'Step size of segmentation in Angstrom'
        feature_set.parameters[inp5] = int(70)
        feature_set.hints[inp5] = 'Overlapping segments are related views according to helical symmetry, i.e. step ' + \
        'size should be a multiple of helical rise (stepsize of 0 corresponds to one central box per helix).'
        feature_set.properties[inp5] = feature_set.Range(0, 2000, 1)
        feature_set.level[inp5]='beginner'
        
        return feature_set


    def set_perturb_step_option(self, feature_set):
        inp6 = 'Perturb step option'
        feature_set.parameters[inp6] = bool(False)
        feature_set.hints[inp6] = 'Perturb the segmentation step between the windowed segments. Takes specified ' + \
        'step size and applies a random shift along the helix between +/- stepsize / 2 . This is ' + \
        'useful to avoid artifacts in the Fourier transforms of class averages.'
        feature_set.level[inp6]='intermediate'
        
        return feature_set
    
    
    def set_invert_option(self, feature_set):
        inp6 = 'Invert option'
        feature_set.parameters[inp6] = bool(True)
        feature_set.hints[inp6] = 'Inversion of image densities for cryo data, i.e. protein becomes white.'
        feature_set.level[inp6]='beginner'
        
        return feature_set


    def set_center_option(self, feature_set):
        inp7 = 'Center option'
        feature_set.parameters[inp7] = bool(False)
        feature_set.hints[inp7] = 'Segments are centered with respect to their helix axis. Use this option with ' + \
        'caution as it will only work for rigid and dense specimens, e.g. tubulur crystals.'
        feature_set.level[inp7]='intermediate'

        return feature_set


    def set_remove_helix_ends_option(self, feature_set):
        inp7 = 'Remove helix ends option'
        feature_set.parameters[inp7] = bool(False)
        feature_set.hints[inp7] = 'Ends of helices are removed by half the segment size. This depends on how you ' + \
        'boxed the helices.'
        feature_set.level[inp7]='intermediate'

        return feature_set


    def set_normalization_option(self, feature_set):
        inp7 = 'Normalization option'
        feature_set.parameters[inp7] = bool(True)
        feature_set.hints[inp7] = 'Segments are normalized with a mean of 0 and standard deviation of 1.'
        feature_set.level[inp7]='intermediate'
        
        return feature_set


    def set_rotation_option(self, feature_set):
        inp8 = 'Rotation option'
        feature_set.parameters[inp8] = bool(True)
        feature_set.hints[inp8] = 'Segments are rotated with helix axis perpendicular to image rows.' 
        feature_set.level[inp8]='intermediate'
        
        return feature_set
    
    
    def set_straightening_option(self, feature_set):
        inp8 = 'Unbending option'
        feature_set.parameters[inp8] = bool(False)
        feature_set.hints[inp8] = 'Segments are computationally unbent or straightened according to coordinates of ' + \
        'helix axis in addition to rotation.'
        if 'Rotation option' in feature_set.parameters:
            feature_set.relatives[inp8]='Rotation option'
        feature_set.level[inp8]='experimental'
        
        return feature_set
    
    
    def set_ctf_correct_option(self, feature_set):
        inp8 = 'CTF correct option'
        feature_set.parameters[inp8] = bool(True)
        feature_set.hints[inp8] = 'Segments are CTF corrected with determined CTF parameters.'
        feature_set.level[inp8]='intermediate'
        
        return feature_set
    

    def set_rows_normalization_option(self, feature_set):
        inp13 = 'Row normalization option'
        feature_set.parameters[inp13] = bool(False)
        feature_set.hints[inp13] = 'Option to normalize micrographs row by row to eliminate artifacts as they occur '+\
        'in Falcon II images or frames if they are not correctly linearized.'
        feature_set.level[inp13]='intermediate'
        
        return feature_set
    
    
    def set_ctf_correct(self, feature_set):
        feature_set = self.set_ctf_correct_option(feature_set)
        feature_set = SegmentCtfApplyPar().set_ctffind_or_ctftilt_choice(feature_set)
        feature_set = SegmentCtfApplyPar().set_ctfconvolve_or_ctfphase_flip_option(feature_set)
        feature_set = SegmentCtfApplyPar().set_astigmatism_option(feature_set)
        
        return feature_set
        
        
    def set_frame_processing_option(self, feature_set):
        inp5 = 'Frame processing option'
        feature_set.parameters[inp5] = bool(False)
        feature_set.hints[inp5] = 'This option will prepare of stack containing frame helix segments from direct ' + \
        'electron detectors and is intended for subsequent helix-based movie processing using \'segmentrefine3d\'. ' + \
        'Prior to this option run \'segmentrefine3d\' using the combined average of all frames. For input of the \'Frame ' + \
        'processing option\' using \'segment\' please provide: 1. \'Micrographs\' as an mrc-stack file ' + \
        '2. \'Segment coordinates\' - use previous spring.db as input instead of pure coordinate files. ' + \
        '3. \'spring.db file\' previous spring.db (same file as 2.) and ' + \
        '4. \'Refinement.db to process\' from your last \'segmentrefine3d\' cycle. This option will generate the ' + \
        'following output: 1. Stack of frame helix segments, ' + \
        '2. spring_frames.db with copies of all segment entries from the previous spring.db and ' + \
        '3. refinement_frames.db with copies of previous orientation parameters. With those output files of the ' + \
        '\'segment\' run you can launch \'segmentrefine3d\' with \'Frame motion correction\''
        
        feature_set.level[inp5]='expert'
         
        return feature_set
     
     
    def set_frame_first_and_last_frame(self, feature_set):
        inp5 = 'First and last frame'
        feature_set.parameters[inp5] = ((0, 6))
        feature_set.hints[inp5] = 'Choose first and last frame to be processed from direct detector movies. ' + \
        'Remember, first frame correspond to frame 0.'
        feature_set.properties[inp5] = feature_set.Range(0, 400, 1)
        feature_set.relatives[inp5] = ('Frame processing option', 'Frame processing option')
        feature_set.level[inp5]='expert'
         
        return feature_set
    
        
    def set_ref_db_frame_processing(self, feature_set):
        inp5 = 'Refinement.db to process'
        feature_set.parameters[inp5] = 'refinement.db'
        feature_set.properties[inp5] = feature_set.file_properties(1, ['db'], 'getFiles')
        feature_set.hints[inp5] = 'Input: refinement.db from previous combined average frame run of segmentrefine3d. '
        feature_set.relatives[inp5] = 'Frame processing option'
        feature_set.level[inp5]='expert'
         
        return feature_set


class SegmentAssign(object):

    def create_link(self, micfile, overlapdir):
        """
        Function to create symbolic link in new directory

        #. Input: micfile (file for wich symbolic link is to be created), overlapdir = new directory for\
            symbolic link location
        #. Output: symbolic link absolute path 
        #. Usage: micabsfile = create_link(micfile, ovelapdir)
        """

        micabsfile = os.path.join(overlapdir, os.path.basename(micfile))
        try:
            os.symlink(micfile, micabsfile)
        except OSError:
            os.remove(micabsfile)
            os.symlink(micfile, micabsfile)

        self.log.ilog('Symbolic link {0} created'.format(micabsfile))

        return micabsfile
    
    
    def reorganize_micrographs_and_coordinate_files_into_separate_directories(self, pair):
        """
        * Function to reorganize micrographs and coordinate files into \
        separate directories micrograph and corresponding segment file into \
        separate subdirectories using symbolic links
        
        #. Input: list of 3-tuples (micrograph file, segment file, overlapping letters)
        #. Output: None
        #. Usage: reorganize_micrographs_and_coordinate_files_into_separate_directories(pair)
        """

        for micfile, segfile, overlap in pair:
            overlapdir = overlap
            try:
                os.mkdir(overlapdir)
                self.log.ilog('Directory {0} created'.format(overlapdir))
            except OSError:
                self.log.ilog('Directory {0} exists'.format(overlapdir))
                pass

            self.create_link(micfile, overlapdir)
            self.create_link(segfile, overlapdir)
            
        return micfile, segfile, overlapdir
        
        
    def assign_micrograph_segment_pairs(self, micrograph_files, coordinate_files):
        """
        * Function to assign pairs of micrographs and segment files based on \
        their common filenames

        #. Input: micrograph_files = list of input micrographs, \
        coordinate_files = list of segment files
        #. Ouput: list of tuples consisting micrograph, corresponding segment \
        file and their common base filename
        #. Usage: assign_micrograph_segment_pairs(micrograph_files, coordinate_files)

        >>> from spring.segment2d.segment import Segment
        >>> s = Segment()
        >>> mics = ['A.tif', 'B.tif', 'C.tif', 'Y.tif']
        >>> boxes = ['A.box', 'B.box', 'C.box', 'Z.box']
        >>> s.assign_micrograph_segment_pairs(mics, boxes)
        [('A.tif', 'A.box', 'A'), ('B.tif', 'B.box', 'B'), ('C.tif', 'C.box', 'C')]
        >>> boxes = ['A_boxes.txt', 'B_boxes.txt', 'C_boxes.txt', 'Z_boxes.txt']
        >>> s.assign_micrograph_segment_pairs(mics, boxes)
        [('A.tif', 'A_boxes.txt', 'A'), ('B.tif', 'B_boxes.txt', 'B'), ('C.tif', 'C_boxes.txt', 'C')]
        """

        if coordinate_files[0].endswith('_boxes.txt'):
            coord_files = dict([(os.path.splitext(os.path.basename(each_coord_file))[0][0:-len('_boxes')], each_coord_file) 
                            for each_coord_file in coordinate_files])
        else:
            coord_files = dict([(os.path.splitext(os.path.basename(each_coord_file))[0], each_coord_file) 
                            for each_coord_file in coordinate_files])

        pair = []
        for micrograph_file in micrograph_files:
            micrograph_base = os.path.splitext(os.path.basename(micrograph_file))[0]
            if coord_files.has_key(micrograph_base):
                common_coord = coord_files[micrograph_base]
                pair.append((micrograph_file, common_coord, micrograph_base))
            else:
                self.log.ilog('Micrograph {0} is ignored could not be matched with a segment file'.\
                              format(micrograph_file))

        log_info = []
        for each_micrograph, each_coord_file, each_overlap in pair:
            log_info += [[os.path.basename(each_micrograph), os.path.basename(each_coord_file), each_overlap]]

        msg = tabulate(log_info, ['micrograph', 'coordinate_file', 'overlapping_string'])
        self.log.ilog('The following micrograph/coordinate pairs have been assigned:\n{0}'.format(msg))

        return pair
    

    def remove_helices_segments_from_tables(self, session, segments, combined_included_segments, segment_table,
    helix_table, red_helix_ids):
        helices = session.query(helix_table).order_by(helix_table.id).all()
 
        SpringDataBase().remove_all_previous_entries(session, helix_table)
        session.commit()
 
        red_helix_id = 0
        hel_columns = SpringDataBase().get_columns_from_table(helix_table)
        helices_to_update = np.zeros(len(helices))
        for each_helix in helices:
            if each_helix.id in list(set(red_helix_ids)):
                data = SpringDataBase().get_data_from_entry(hel_columns, each_helix)
                red_helix_id += 1
                data['id'] = red_helix_id
                session.add(helix_table(**data))
                helices_to_update[each_helix.id - 1] = red_helix_id

        SpringDataBase().remove_all_previous_entries(session, segment_table)
        session.commit()

        red_stack_id = 0
        seg_columns = SpringDataBase().get_columns_from_table(segment_table)
        for each_segment in segments:
            if each_segment.stack_id in combined_included_segments:
                data = SpringDataBase().get_data_from_entry(seg_columns, each_segment)
                data['stack_id'] = red_stack_id
                if data.has_key('helix_id'):
                    data['helix_id'] = helices_to_update[red_helix_ids[red_stack_id] - 1]
                red_stack_id += 1
                data['id'] = red_stack_id
                session.add(segment_table(**data))
        
        session.commit()


    def perform_helix_based_segment_selection_incl_segment_helix_removal_from_springdb(self, db_name):
        if self.straightness_selection or self.mics_selection or self.helices_selection or self.defocus_selection \
        or self.astigmatism_selection:
            shutil.copy(db_name, 'spring.db')
            db_name = 'spring.db'
            session = SpringDataBase().setup_sqlite_db(base, db_name)
            segments = session.query(SegmentTable).order_by(SegmentTable.stack_id).all()

            self.curvature_range, self.ccc_layer_range = SegmentSelect().convert_curvature_ccc_layer_range(db_name,
            self.straightness_selection, self.curvature_range_perc)

            included_segments_mics, included_segments_helices, included_segments_curve, included_segments_defocus, \
            included_segments_astigmatism, excluded_mics_count, excluded_helix_count, excluded_curvature_count, \
            excluded_defocus_count, excluded_astig_count = \
            SegmentSelect().filter_by_properties_that_keep_helices_together(self, session, segments)

            combined_included_segments = list(set(range(len(segments))).intersection(included_segments_mics, 
                    included_segments_helices, included_segments_curve, included_segments_defocus, 
                    included_segments_astigmatism))

            red_helix_ids = [each_segment.helix_id for each_segment in segments \
                             if each_segment.stack_id in combined_included_segments]

            self.remove_helices_segments_from_tables(session, segments, combined_included_segments, SegmentTable,
            HelixTable, red_helix_ids)
            
            if self.frame_option:
                ref_db = 'refinement.db'
                shutil.copy(self.ref_db, ref_db)
                self.ref_db = ref_db
                ref_session = SpringDataBase().setup_sqlite_db(refine_base, ref_db)
                
                ref_segments = ref_session.query(RefinementCycleSegmentTable).\
                                        order_by(RefinementCycleSegmentTable.stack_id).all()
                
                self.remove_helices_segments_from_tables(ref_session, ref_segments, combined_included_segments,
                RefinementCycleSegmentTable, RefinementCycleHelixTable, red_helix_ids)
            
        return db_name


    def assign_reorganize(self, micrograph_files, coordinate_files):
        """
        * Function to match micrographs and corresponding segment file 

        #. Input: list of input micrographs, list of segment files
        #. Output: assigned pair of micrograph and segments
        #. Usage: assign_reorganize(micrograph_files, coordinate_files)
        """
        self.log.fcttolog()

        if coordinate_files[0].endswith('db'):
            db_name = coordinate_files[0]
            db_name = self.perform_helix_based_segment_selection_incl_segment_helix_removal_from_springdb(db_name)
                
            session = SpringDataBase().setup_sqlite_db(base, db_name)
            helices = session.query(HelixTable).order_by(HelixTable.id).all()
        
            mic_ids = list(set([each_helix.mic_id for each_helix in helices]))

            mics = [session.query(CtfMicrographTable).get(each_mic_id) for each_mic_id in mic_ids]

            if os.path.exists(os.path.join(mics[0].dirname, mics[0].micrograph_name)):
                pair = [(str(os.path.join(each_mic.dirname, each_mic.micrograph_name)), db_name, 
                         os.path.splitext(os.path.basename(each_mic.micrograph_name))[0]) 
                        for each_mic in mics]
            else:
                base_name_mic_files = [os.path.splitext(os.path.basename(each_mic_name))[0] 
                                       for each_mic_name in micrograph_files]
                
                mic_dict = {}
                for each_file in micrograph_files:
                    each_file_base = os.path.basename(each_file)
                    for each_base_name in base_name_mic_files:
                        if each_file_base.startswith(each_base_name):
                            mic_dict[each_base_name]=each_file

                pair = [(mic_dict[str(os.path.splitext(each_mic.micrograph_name)[0])], db_name, 
                os.path.splitext(os.path.basename(each_mic.micrograph_name))[0]) for each_mic in mics]
        else:
            pair = self.assign_micrograph_segment_pairs(micrograph_files, coordinate_files)

        if pair == []:
            errstring = 'No micrographs could be matched with segment files. Nothing to be done. Abort {0}'.\
            format(os.path.basename(__file__).split(os.extsep)[0])
            self.log.errlog(errstring)
            raise IOError(errstring)

        self.reorganize_micrographs_and_coordinate_files_into_separate_directories(pair)
        
        return pair
    

class SegmentPreparation(Micrograph, SegmentAssign):
    """
    * Class that holds functions for extracting segments from micrographs

    * __init__ Function to interpret multi-input parameters

    """
    def __init__(self, parset = None):
        self.log = Logger()
        if parset is not None:
            self.feature_set = parset
            p = self.feature_set.parameters

            self.infile = p['Micrographs']
            self.outfile = p['Image output stack']
            self.spring_db_option = p['Spring database option']
            self.spring_path = p['spring.db file']
            self.spring_path = os.path.abspath(self.spring_path)
            
            self.pixelsize = float(p['Pixel size in Angstrom'])
            self.define_segment_or_window_parameters(p)
            self.define_ctf_parameters(p)
            
            self.invertoption=p['Invert option']
            self.centeroption=False
            self.normoption=p['Normalization option']
            self.row_normalization=p['Row normalization option']
            self.binoption=p['Binning option']
            self.binfactor=p['Binning factor']
            if self.binfactor == 1 and self.binoption is True:
                self.binoption = False

            self = SegmentSelect().define_mics_and_helices_selection(self, p)
            self = SegmentSelect().define_straightness_selection(self, p)
            self = SegmentSelect().define_defocus_and_astigmatism_selection(self, p)

            self.segsizepix = self.determine_boxsize_closest_to_fast_values(self.window_size / self.pixelsize, ten=True)
            if self.ctfcorrect_option:
                ctf_window_size = 1.3 * self.window_size / self.pixelsize
                self.window_sizepix = self.determine_boxsize_closest_to_fast_values(ctf_window_size)
            else:
                self.window_sizepix = self.segsizepix

            self.frame_option = p['Frame processing option']
            self.frame_range = p['First and last frame']
            self.spring_db_frames = 'spring_{0}-{1}frames.db'.format(self.frame_range[0], self.frame_range[-1])
            self.ref_db = p['Refinement.db to process']
            self.ref_db_frames = 'refinement_{0}-{1}frames.db'.format(self.frame_range[0], self.frame_range[-1])
            
            self.micrograph_files = Features().convert_list_of_files_from_entry_string(self.infile)
            self.coordinate_files = Features().convert_list_of_files_from_entry_string(self.coordinate_files_entry)
            

    def define_segment_or_window_parameters(self, p):
        self.coordinate_files_entry = p['Segment coordinates']
        self.segment_size = p['Segment size in Angstrom']
        
        self.remove_ends=p['Remove helix ends option']
        if Features().convert_list_of_files_from_entry_string(self.coordinate_files_entry)[0].endswith('db'):
            self.remove_ends = False
        
        self.stepsize = p['Step size of segmentation in Angstrom']
        if self.stepsize == 0:
            self.remove_ends = False
            
        self.perturb_step = p['Perturb step option']
        self.window_size = self.segment_size + 1.2 * self.stepsize
        self.helixwidth = p['Estimated helix width in Angstrom']
        self.helixwidthpix = int(round(self.helixwidth/self.pixelsize))
        
        self.rotoption=p['Rotation option']
        self.unbending=p['Unbending option']

        self.averaging_option=False
        self.averaging_distance=10 * self.stepsize#p['Local averaging distance in Angstrom']
        
        self.mpi_option = p['MPI option']
        self.cpu_count = p['Number of CPUs']
        if self.cpu_count == 1:
            self.mpi_option = False
        self.temppath=p['Temporary directory']
            

    def define_ctf_parameters(self, p):
        self.ctfcorrect_option = p['CTF correct option']
        self = SegmentCtfApply().define_ctf_correction_parameters(self, p)


    def get_strip_frame(self, first_frame):
        if first_frame == 0:
            first_frame_strip = ''
        else:
            first_frame_strip = '{0}'.format(first_frame)
        
        return first_frame_strip
        

    def get_message_for_frame_missing(self, first_frame, first_str, mics):
        msg = 'Specified {0} frame {1} cannot be found in pool of provided micrographs ({2}). '.\
        format(first_str, first_frame, ', '.join(mics)) + \
        'Double-check existance of frame or correct {0} frame entry.\n'.format(first_str)

        return msg


    def check_whether_provided_frame_number_matches_provided_micrographs(self, micrographs, input_mics, frame_range):
        """
        >>> from spring.segment2d.segment import Segment
        >>> s = Segment()
        >>> mics = ['test.hdf', 'tost.hdf']
        >>> input_mics = ['test_{0:03}.hdf'.format(each) for each in list(range(5))]
        >>> s.check_whether_provided_frame_number_matches_provided_micrographs(mics, input_mics, (0, 4))
        [('test.hdf', ['test_000.hdf', 'test_001.hdf', 'test_002.hdf', 'test_003.hdf', 'test_004.hdf'])]
        >>> s.check_whether_provided_frame_number_matches_provided_micrographs(mics, input_mics, (2, 4))
        [('test.hdf', ['test_002.hdf', 'test_003.hdf', 'test_004.hdf'])]
        >>> s.check_whether_provided_frame_number_matches_provided_micrographs(mics, input_mics, (2, 7)) #doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: Specified last frame 7 cannot be found in pool of provided 
        micrographs (test_000.hdf, test_001.hdf, test_002.hdf, test_003.hdf, 
        test_004.hdf). Double-check existance of frame or correct last frame entry.
        """
        first_frame, last_frame = frame_range
        first_frame_strip = self.get_strip_frame(first_frame)
        last_frame_strip = self.get_strip_frame(last_frame)

        assigned_inp_mics = []
        expected_pattern = set(['{0}'.format(each_id) if each_id != 0 else '' 
                             for each_id in list(range(first_frame, last_frame + 1))])
            
        for each_mic in micrographs:
            micrograph_base = os.path.splitext(os.path.basename(each_mic))[0]

            common_inp = [os.path.basename(each_inp) for each_inp in input_mics if micrograph_base in each_inp]
            common_full_path_inp = [each_inp for each_inp in input_mics if micrograph_base in each_inp]

            mics = [each_full_inp for each_inp, each_full_inp in zip(common_inp, common_full_path_inp) 
                    if os.path.splitext(each_inp)[0].startswith(micrograph_base)]

            diff = [os.path.splitext(each_inp)[0][len(micrograph_base):] for each_inp in common_inp 
                    if os.path.splitext(each_inp)[0].startswith(micrograph_base)]

            diff_strip = [each_diff.strip('_').lstrip('0') for each_diff in diff]
            
            assigned_mics = dict(zip(diff_strip, mics))
            if mics != []:
                msg = ''
                if not assigned_mics.has_key(first_frame_strip):
                    msg += self.get_message_for_frame_missing(first_frame, 'first', mics)
                elif not assigned_mics.has_key(last_frame_strip):
                    msg += self.get_message_for_frame_missing(last_frame, 'last', mics)
                if not assigned_mics.has_key(first_frame_strip) or not assigned_mics.has_key(last_frame_strip):
                    raise ValueError(msg)

            overlap = expected_pattern.intersection(diff_strip)
            if len(overlap) > 0:
                common_inp_full_path = [assigned_mics[each_overlap] for each_overlap in overlap]
                common_inp_full_path.sort()
                assigned_inp_mics.append((os.path.basename(each_mic), common_inp_full_path))

        return assigned_inp_mics

    
    def check_whether_provided_frame_number_matches_provided_micrographs_old(self, micrographs, input_mics, frame_count):
        """
        >>> from spring.segment2d.segment import Segment
        >>> s = Segment()
        >>> mics = ['test.hdf', 'tost.hdf']
        >>> input_mics = ['test_{0:03}.hdf'.format(each) for each in list(range(5))]
        >>> s.check_whether_provided_frame_number_matches_provided_micrographs_old(mics, input_mics, 5)
        [('test.hdf', ['test_000.hdf', 'test_001.hdf', 'test_002.hdf', 'test_003.hdf', 'test_004.hdf'])]
        >>> input_mics = ['test_{0:02}.hdf'.format(each) for each in list(range(5))]
        >>> s.check_whether_provided_frame_number_matches_provided_micrographs_old(mics, input_mics, 5)
        [('test.hdf', ['test_00.hdf', 'test_01.hdf', 'test_02.hdf', 'test_03.hdf', 'test_04.hdf'])]
        >>> input_mics = ['test_{0:02}.hdf'.format(each) for each in list(range(5, 10))]
        >>> s.check_whether_provided_frame_number_matches_provided_micrographs_old(mics, input_mics, 5)
        [('test.hdf', ['test_05.hdf', 'test_06.hdf', 'test_07.hdf', 'test_08.hdf', 'test_09.hdf'])]
        >>> s.check_whether_provided_frame_number_matches_provided_micrographs_old(mics, input_mics, 6) #doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: The number of provided frames (=6) from test_05.hdf, 
        test_06.hdf, test_07.hdf, test_08.hdf, test_09.hdf does not match the 
        'Processing frame number' (=6) specified. Please, change input 
        files of frames or the 'Processing frame number'.
        >>> input_mics = ['test_ttt.hdf'.format(each) for each in list(range(5))]
        >>> s.check_whether_provided_frame_number_matches_provided_micrographs_old(mics, input_mics, 4) #doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: The provided frames (test_ttt.hdf, test_ttt.hdf, 
        test_ttt.hdf, test_ttt.hdf, test_ttt.hdf) can not be 
        assigned because they do not follow the pattern: test.hdf - 
        test_000.hdf, test_001.hdf, test_002.hdf, test_003.hdf, test_004.hdf.
        Please rename to make frames assignable to previous micrograph.
        """
        assigned_inp_mics = []
        for each_mic in micrographs:
            micrograph_base = os.path.splitext(os.path.basename(each_mic))[0]

            common_inp = [os.path.basename(each_inp) for each_inp in input_mics if micrograph_base in each_inp]
            diff = [os.path.splitext(each_inp)[0][len(micrograph_base):] for each_inp in common_inp 
                    if os.path.splitext(each_inp)[0].startswith(micrograph_base)]
            diff_strip = [each_diff.strip('_').strip('0') for each_diff in diff]
            
            expected_patterns = []
            for each_first_frame in list(range(400)):
                expected_patterns.append(['{0}'.format(each_id) if each_id != 0 else '' 
                for each_id in list(range(each_first_frame, each_first_frame + len(diff_strip)))])
            
            if diff_strip in expected_patterns:
                common_inp_full_path = [each_inp for each_inp in input_mics if micrograph_base in each_inp]
                if len(diff) > 0:
                    if len(diff) != frame_count:
                        msg = 'The number of provided frames (={0}) from {1} '.format(frame_count, ', '.join(common_inp)) + \
                        'does not match the \'Processing frame number\' (={0}) specified. '.format(frame_count) + \
                        'Please, change input files of frames or the \'Processing frame number\'.'
                        raise ValueError(msg)
                    else:
                        assigned_inp_mics.append((each_mic, common_inp_full_path))
            else:
                expected_pattern_underscore = ['_{0:03}'.format(each_id) for each_id in list(range(len(diff)))]
                micrograph_ext = os.path.splitext(os.path.basename(each_mic))[-1]
                expected_mics = [micrograph_base + each_pattern + micrograph_ext 
                                 for each_pattern in expected_pattern_underscore]
 
                msg = 'The provided frames ({0}) can not be assigned because '.format(', '.join(common_inp)) + \
                'they do not follow the pattern: {0} - {1}. '.format(os.path.basename(each_mic), ', '.join(expected_mics) ) + \
                'Please rename to make frames assignable to previous micrograph.'
 
                raise ValueError(msg)

        return assigned_inp_mics

    
    def check_whether_spring_db_entries_match_provided_micrographs(self):
        if not self.coordinate_files[0].endswith('db'):
            msg = 'Please provide coordinates from previous spring.db that need to be frame processed.'
            raise ValueError(msg)
        else:
            spring_db = self.coordinate_files[0]

        session = SpringDataBase().setup_sqlite_db(base, spring_db)
        helices = session.query(HelixTable).order_by(HelixTable.id).all()
        
        mic_ids = []
        [mic_ids.append(each_helix.mic_id) for each_helix in helices if each_helix.mic_id not in mic_ids]

        mics = session.query(CtfMicrographTable).all()
        
        demanded_frame_count = self.frame_range[-1] - self.frame_range[0]
        micrographs = []
        for each_mic in mics:
#             mic_name = os.path.join(each_mic.dirname, each_mic.micrograph_name)
            mic_name = each_mic.micrograph_name
            mic_base = os.path.splitext(each_mic.micrograph_name)[0]
            micrographs.append(str(mic_name))
            search_str_mics = ' '.join(self.micrograph_files)
            if search_str_mics.find(mic_base) < 0:
                msg = 'Micrograph {0} could not be found in provided micrographs. Double-check existence.'.format(mic_name)
                raise ValueError(msg)
            if search_str_mics.find(mic_base) < demanded_frame_count:
                msg = 'The processing frame number does not match the number of provided micrographs for ' + \
                '{0}. Double check.'.format(mic_name)
                raise ValueError(msg)

        return micrographs

    
    def validate_input(self):
        if 0 < self.stepsize < self.pixelsize: 
            error_message = 'Segmentation step size can not be smaller than pixel size.'
            raise ValueError(error_message)
        
        coord_ending = list(set([os.path.splitext(each_file)[-1] for each_file in self.coordinate_files]))
        if not coord_ending[0].endswith('db'):
            if self.straightness_selection or self.mics_selection or self.helices_selection or \
            self.defocus_selection or self.astigmatism_selection:
                err_msg = 'You have specified selection criteria, please provide the coordinates with a ' + \
                'spring.db from previous segment or segmentrefine3d runs. This spring.db will be used for ' + \
                'assessing the criteria.'
                raise ValueError(err_msg)
                
        if self.frame_option:
            first_frame, last_frame = self.frame_range
            if last_frame < first_frame:
                err_msg = 'The specified first frame {0} is greater than '.format(first_frame) + \
                'the last frame {0}. Please make first frame smaller or equal to last frame.'.format(last_frame)
                raise ValueError(err_msg)
            
            if not coord_ending[0].endswith('db'):
                err_msg = 'As you specified the \'Frame processing option\', please provide coordinates by the ' + \
                'the previously run spring.db instead of the original coordinate files.'
                raise ValueError(err_msg)

            ending_mics = set([os.path.splitext(each_mic)[-1] for each_mic in self.micrograph_files])
            if list(ending_mics)[0].endswith('mrcs'):
                frames = EMData()
                frames.read_image(self.micrograph_files[0])
                frame_count = frames.get_zsize()
                if frame_count < last_frame:
                    err_msg = 'Specified frame {0} as last frame cannot be found in provided '.format(last_frame) + \
                    'mrcs stack file. {0} only contains {1} frames. '.format(self.micrograph_files[0], frame_count) + \
                    'Please, adjust the last frame input value to maximum of {0} accordingly '.format(frame_count) + \
                    '(first frame is 0).'
                    raise ValueError(err_msg)

            if len(ending_mics) == 1 and list(ending_mics)[0].endswith('mrcs'):
                assigned_mics = []
                for each_mic in self.micrograph_files:
                    mic_frames = [os.path.splitext(each_mic)[0] + '@{0}{1}mrcs'.format(each_frame, os.extsep) 
                                  for each_frame in range(first_frame, last_frame + 1)]
                    assigned_mics.append((os.path.basename(each_mic), mic_frames))
            else:
                mics = self.check_whether_spring_db_entries_match_provided_micrographs()

                assigned_mics = self.check_whether_provided_frame_number_matches_provided_micrographs(mics,
                self.micrograph_files, self.frame_range)

            source_session = SpringDataBase().setup_sqlite_db(base, self.spring_path)
            new_session = SpringDataBase().setup_sqlite_db(base, self.spring_db_frames)

            new_session = self.copy_micrograph_info_on_ctffind_and_ctftilt(assigned_mics, source_session, new_session)
        else:
            assigned_mics = None
        
        return assigned_mics

        
    def determine_boxsize_closest_to_fast_values(self, boxsize_ori, ten=False):
        """
        # Values taken from http://blake.bcm.edu/emanwiki/EMAN2/BoxSize
        >>> from spring.segment2d.segment import Segment
        >>> s = Segment()
        >>> s.determine_boxsize_closest_to_fast_values(333)
        352
        >>> s.determine_boxsize_closest_to_fast_values(333, ten=True)
        360
        """
        fast_size = [32, 33, 35, 40, 44, 48, 52, 64, 66, 72, 84, 100, 104, 112, 128, 130, 132, 140, 150, 160, 168, 180,
        182, 192, 196, 220, 224, 240, 256, 260, 288, 300, 320, 324, 330, 352, 360, 384, 416, 420, 440, 448, 450, 480,
        512, 540, 576, 600, 625, 640, 648, 675, 720, 729, 750, 768, 800, 810, 864, 900, 960, 972, 1000, 1024, 1080,
        1125, 1152, 1200, 1215, 1250, 1280, 1296, 1350, 1440, 1458, 1500, 1536, 1600, 1620, 1728, 1800, 1875, 1920,
        1944, 2000, 2025, 2048, 2160, 2187, 2250, 2304, 2400, 2430, 2500, 2560, 2592, 2700, 2880, 2916, 3000, 3072,
        3125, 3200, 3240, 3375, 3456, 3600, 3645, 3750, 3840, 3888, 4000, 4050, 4320, 4374, 4500, 4608, 4800, 4860,
        5000, 5120, 5184, 5400, 5625, 5760, 5832, 6000, 6075, 6144, 6250, 6400, 6480, 6750, 6912, 7200, 7290, 7500,
        7680, 7776, 8000, 8100]
        
        if ten:
            fast_size = [each_size for each_size in fast_size if (each_size) % 10 == 0]

        fast_size_id = np.argmin(np.abs(boxsize_ori - np.array(fast_size)))
        fast_box = fast_size[fast_size_id]
        if fast_box < boxsize_ori:
            fast_box = fast_size[fast_size_id + 1]
        
        return fast_box
        

class SegmentPreparationFileRead(SegmentPreparation):
    def read_coordinate_lines_from_eman_format_and_separate_helices(self, box_coord):
        """
        >>> lines = ['50\\t50\\t100\\t100\\t-1\\n', '40\\t40\\t100\\t100\\t0\\n']
        >>> lines += ['30\\t30\\t100\\t100\\t-2\\n', '20\\t20\\t100\\t100\\t-1\\n',]
        >>> box_coord = lines + ['10\\t10\\t100\\t100\\t-2\\n']
        >>> from spring.segment2d.segment import Segment
        >>> s = Segment()
        >>> s.read_coordinate_lines_from_eman_format_and_separate_helices(box_coord) #doctest: +NORMALIZE_WHITESPACE
        ['50\\t50\\t100\\t100\\t-1\\n40\\t40\\t100\\t100\\t0\\n30\\t30\\t100\\t100\\t-2\\n', 
        '20\\t20\\t100\\t100\\t-1\\n10\\t10\\t100\\t100\\t-2\\n']
        """
        helixcoord = []
        helix = ''
        for each_file_line in box_coord:
            helix = helix + each_file_line
            if each_file_line.endswith('-2\n'):
                helixcoord.append(helix)
                helix = ''
        
        if helixcoord == []:
            helixcoord.append(helix)
            
        return helixcoord


    def add_start_or_end_flag(self, bsoft_coord, id_pos, helixcoord, helix, previous_helix_id, each_index,
    current_helix_id):
        next_line = bsoft_coord[each_index + 1]
        if next_line != '\n':
            next_helix_id = next_line.split()[id_pos]
            if previous_helix_id != current_helix_id:
                helix = helix + '\t-1\n'
                previous_helix_id = current_helix_id
            elif next_helix_id != current_helix_id:
                helix = helix + '\t-2\n'
                helixcoord.append(helix)
                helix = ''
            else:
                helix = helix + '\t0\n'
                
        return helix, previous_helix_id


    def append_final_helix(self, helixcoord, helix):
        helix = helix + '\t-2\n'
        helixcoord.append(helix)
        coordinate_lines = False
        
        return helixcoord, coordinate_lines
    

    def read_coordinate_lines_from_bsoft_particle_format_and_convert_to_eman_format(self, bsoft_coord, segment_size=40.0):
        """
        >>> bsoft = [ '_particle.select\\n']
        >>> ln1 = '1    1       0  1.0000  366.00 2058.00    0.00  20.000  20.0'
        >>> ln1 += '00   0.000  1.0000  0.0000  0.0000   34.79  1.0000    1\\n'
        >>> ln2 = '2    1       0  1.0000  382.43 2046.59    0.00  20.000  20.0'
        >>> ln2 += '00   0.000  1.0000  0.0000  0.0000   34.79  1.0000    1\\n'
        >>> ln3 = '3    2       0  1.0000 1326.17 2107.80    0.00  20.000  20.0'
        >>> ln3 += '00   0.000  1.0000  0.0000  0.0000  -54.08  1.0000    2\\n'
        >>> ln4 = '4    2       0  1.0000 1326.17 2107.80    0.00  20.000  20.0'
        >>> ln4 += '00   0.000  1.0000  0.0000  0.0000  -54.08  5.0000    2\\n'
        >>> ln5 = '5    2       0  1.0000 1326.17 2107.80    0.00  20.000  20.0'
        >>> ln5 += '00   0.000  1.0000  0.0000  0.0000  -54.08  5.0000    2\\n'
        >>> bsoft += [ln1, ln2, ln3, ln4, ln5, '\\n', '\\n', 'loop_\\n']
        >>> from spring.segment2d.segment import Segment
        >>> s = Segment()
        >>> s.read_coordinate_lines_from_bsoft_particle_format_and_convert_to_eman_format(bsoft) #doctest: +NORMALIZE_WHITESPACE
        ['346.0\\t2038.0\\t40.0\\t40.0\\t-1\\n362.43\\t2026.59\\t40.0\\t40.0\\t-2\\n', 
        '1306.17\\t2087.8\\t40.0\\t40.0\\t-1\\n1306.17\\t2087.8\\t40.0\\t40.0\\t0\\n1306.17\\t2087.8\\t40.0\\t40.0\\t-2\\n']
        >>> bsoft = [ '_filament.z\\n','      1      1  366.00 2058.00    0.00\\n']
        >>> bsoft += ['      1      2  461.00 1992.00    0.00\\n']
        >>> bsoft += ['      2      1  553.00 1905.00    0.00\\n']
        >>> bsoft += ['      3      1  641.00 1838.00    0.00\\n']
        >>> bsoft += ['      3      2  707.33 1800.67    0.00\\n', '\\n']
        >>> s.read_coordinate_lines_from_bsoft_particle_format_and_convert_to_eman_format(bsoft)
        []
        """
        helixcoord = []
        helix = ''
        coordinate_lines = False
        previous_helix_id = '0'
        for each_index, each_file_line in enumerate(bsoft_coord):
            if each_file_line.startswith('_particle.select'):
                coordinate_lines = True
            if coordinate_lines and len(each_file_line.split()) == 16:
                current_helix_id = (each_file_line.split()[-1])
                each_coord_line = '{x}\t{y}\t{box_x}\t{box_y}'.format(x=float(each_file_line.split()[4]) - segment_size/2,
                y=float(each_file_line.split()[5]) - segment_size/2, box_x=segment_size, box_y=segment_size)
                                                                                      
                helix = helix + each_coord_line
                
                helix, previous_helix_id = self.add_start_or_end_flag(bsoft_coord, -1, helixcoord, helix,
                previous_helix_id, each_index, current_helix_id)
                
            if each_file_line.startswith('loop_') and coordinate_lines:
                helixcoord, coordinate_lines= self.append_final_helix(helixcoord, helix)
        
        return helixcoord
    
    
    def read_coordinate_lines_from_bsoft_filament_format_and_convert_to_eman_format(self, bsoft_coord,
    segment_size=40.0):
        """
        >>> bsoft =  [ '_filament.z\\n']
        >>> bsoft += ['      1      1  366.00 2058.00    0.00\\n']
        >>> bsoft += ['      1      2  461.00 1992.00    0.00\\n']
        >>> bsoft += ['      1      3  553.00 1905.00    0.00\\n']
        >>> bsoft += ['      2      1  641.00 1838.00    0.00\\n']
        >>> bsoft += ['      2      2  707.33 1800.67    0.00\\n', '\\n']
        >>> from spring.segment2d.segment import Segment
        >>> s = Segment()
        >>> s.read_coordinate_lines_from_bsoft_filament_format_and_convert_to_eman_format(bsoft) #doctest: +NORMALIZE_WHITESPACE
        ['346.0\\t2038.0\\t40.0\\t40.0\\t-1\\n441.0\\t1972.0\\t40.0\\t40.0\\t0\\n533.0\\t1885.0\\t40.0\\t40.0\\t-2\\n', 
        '621.0\\t1818.0\\t40.0\\t40.0\\t-1\\n687.33\\t1780.67\\t40.0\\t40.0\\t-2\\n']
        >>> bsoft = ['_filament.id\\n', '_filament.node_id\\n']
        >>> bsoft += ['_filament.id\\n', '_filament.node_id\\n', '_filament.x\\n']
        >>> bsoft += ['_filament.y\\n', '_filament.z\\n'] 
        >>> bsoft += ['      1      1  100.00  100.00    0.00\\n']
        >>> bsoft += ['      1      2   60.00   60.00    0.00\\n', '\\n', '\\n']
        >>> s.read_coordinate_lines_from_bsoft_filament_format_and_convert_to_eman_format(bsoft)
        ['80.0\\t80.0\\t40.0\\t40.0\\t-1\\n40.0\\t40.0\\t40.0\\t40.0\\t-2\\n']
        """
        helixcoord = []
        helix = ''
        coordinate_lines = False
        previous_helix_id = '0'
        for each_index, each_file_line in enumerate(bsoft_coord):
            if each_file_line.startswith('_filament.z'):
                coordinate_lines = True
            if coordinate_lines and len(each_file_line.split()) == 5:
                current_helix_id = (each_file_line.split()[0])
                each_coord_line = '{x}\t{y}\t{box_x}\t{box_y}'.format(x=float(each_file_line.split()[2]) - segment_size/2,
                y=float(each_file_line.split()[3]) - segment_size/2, box_x=segment_size, box_y=segment_size)
                                                                                      
                helix = helix + each_coord_line
                
                helix, previous_helix_id = self.add_start_or_end_flag(bsoft_coord, 0, helixcoord, helix,
                previous_helix_id, each_index, current_helix_id)
                
            if each_file_line == '\n' and coordinate_lines:
                helixcoord, coordinate_lines = self.append_final_helix(helixcoord, helix)
        
        return helixcoord
    


#!/usr/bin/env python
"""
Test module to check segmentexam
"""
from EMAN2 import EMNumPy
from fundamentals import image_decimate
from multiprocessing import cpu_count
from spring.csinfrastr.csproductivity import TestingDirectory
from spring.segment2d.segmentexam import SegmentExam, SegmentExamPar
from spring.tests.csinfrastr.test_csreadinput import EndToEndTest
from spring.tests.segment2d.test_segmentselect import TestSegmentSelect
from spring.tests.segment3d.test_segmentrefine3d import TestSegmentRefine3dSetupForOthers
from utilities import model_gauss_noise
import numpy as np
import os
import shutil


class HelixProjectionPreparation(object):
    def prepare_helix_projection(self):
        """
        * Function to generate a projection of a dummy helix
        """
        self.pixelsize = float(5)
        self.width_of_helix_in_pixel = self.width_of_helix_in_angstrom/self.pixelsize
        self.micrograph_size_in_angstrom = 1000
        
        self.make_empty_segment()
        self.compute_intersect_sine_waves()
        self.add_intersects_to_segment_image()
        self.append_helix_segments_to_stack()
        
    def make_empty_segment(self):
        self.decimation_factor_for_better_layer_lines = 10
        
        self.temporary_segment_size_in_pixels = \
        self.decimation_factor_for_better_layer_lines*self.micrograph_size_in_angstrom/self.pixelsize
        
        self.helix_segment = np.zeros((self.temporary_segment_size_in_pixels, self.temporary_segment_size_in_pixels))

    def compute_intersect_sine_waves(self):
        self.pixels = np.arange(0, self.temporary_segment_size_in_pixels, 0.01)
        
        self.frequency_one = 9
        self.frequency_two = self.frequency_one - 5
        self.sine_wave = []
        for frequency in [self.frequency_one, self.frequency_two]:
            self.sine_wave.append(np.round(self.width_of_helix_in_pixel*2*np.sin(frequency*self.pixels) + \
            self.temporary_segment_size_in_pixels/2, 1))

        self.sine_wave_one = self.sine_wave[0]
        self.sine_wave_two = self.sine_wave[1]

        self.intersect = np.where(np.rint(self.sine_wave_one) == np.rint(self.sine_wave_two))[0]

    def add_intersects_to_segment_image(self):
        for point in self.intersect:
            row = np.rint(self.pixels[point])
            column = np.rint(self.sine_wave_one[point])
            if row < self.temporary_segment_size_in_pixels and column < self.temporary_segment_size_in_pixels:
                self.helix_segment[row][column] = 1

        self.helix_segment_img = EMNumPy.numpy2em(np.copy(self.helix_segment))

        self.helix_segment_img = image_decimate(self.helix_segment_img, self.decimation_factor_for_better_layer_lines,
        fit_to_fft=False)

        self.helix_segment_img.process_inplace('normalize')
        self.helix_segment_img *= 0.1
        
        return self.helix_segment_img
    

    def append_helix_segments_to_stack(self, number=60):
        for segment in list(range(number)):
            self.helix_micrograph = self.helix_segment_img.copy()
            self.add_noise_to_helix_micrograph()
            self.helix_micrograph.append_image('test_segments.hdf')

    def add_noise_to_helix_micrograph(self):
        self.noise = model_gauss_noise(0.1, self.helix_micrograph.get_xsize(), self.helix_micrograph.get_ysize())
        self.helix_micrograph += self.noise

class TestSegmentExam(HelixProjectionPreparation, SegmentExam):
    def setup_segmentrefine3d(self):
        self.tsr3d = TestSegmentRefine3dSetupForOthers()
        self.tsr3d.setup_sr3d()
        os.remove(self.tsr3d.input_stack)

    def setup(self):
        self.testingdir = TestingDirectory(os.path.splitext(os.path.basename(__file__))[0])
        self.testingdir.create()

        self.feature_set = SegmentExamPar()
        self.width_of_helix_in_angstrom = 100
        self.pixelsize = float(5)
        self.width_of_helix_in_pixel = self.width_of_helix_in_angstrom/self.pixelsize
        self.micrograph_size_in_angstrom = 1000
        
        self.prepare_helix_projection()
        self.feature_set.parameters['Image input stack']='test_segments.hdf'
        self.feature_set.parameters['Diagnostic plot']='test_segmentexam.png'
        
        self.feature_set.parameters['spring.db file']=os.path.join(os.pardir, 'spring.db')
            
        self.feature_set.parameters['Power spectrum output image']='test_power.hdf'
        self.feature_set.parameters['Enhanced power spectrum option']=True
        self.feature_set.parameters['Enhanced power spectrum output image']='test_power_enhanced.hdf'
        self.feature_set.parameters['Pixel size in Angstrom']=float(self.pixelsize)
        self.feature_set.parameters['Estimated helix width in Angstrom']=int(1.1*self.width_of_helix_in_angstrom)
        self.feature_set.parameters['Power spectrum resolution cutoff in 1/Angstrom']=float(1/(self.pixelsize*4))
        self.feature_set.parameters['Reference power spectrum']=False
        self.feature_set.parameters['Power spectrum input image']='ref_power.hdf'
        
        self.feature_set.parameters['Compute layer-line correlation option']=False
        self.feature_set.parameters['Layer-line region in 1/Angstrom']=((0.03, 0.04))

        self.tsegsel = TestSegmentSelect()
        self.tsegsel.test_segment_file = 'test.dat'
        self.feature_set = self.tsegsel.setup_selection_criteria_from_segment_table(self.feature_set)
        self.tsegsel.prepare_segment_file()

        self.setup_segmentrefine3d()

        self.feature_set.parameters['MPI option']=False
        self.feature_set.parameters['Number of CPUs']=min(8, cpu_count())
        self.feature_set.parameters['Temporary directory']=os.path.abspath(os.curdir)

        super(TestSegmentExam, self).__init__(self.feature_set)


    def check_width_of_helix(self):
        assert round(self.width_of_helix_in_angstrom, -2) == round(np.mean(self.widths), -2) or \
            round(self.width_of_helix_in_angstrom - 100, -2) == round(np.mean(self.widths), -2)

        
    def teardown(self):
        os.remove(os.path.join(os.pardir, 'spring.db'))
        os.remove('spring.db')
        os.remove(self.tsegsel.test_segment_file)
        os.remove('test_segments.hdf')
        os.remove('test_segmentexam.png')
        os.remove('test_power.hdf')
        os.remove('test_power_enhanced.hdf')

        self.testingdir.remove()

        
class TestSegmentExamMain(TestSegmentExam):
    def do_test_case_se1(self):
        self.add_up_power_spectra()
        self.check_width_of_helix()


class TestSegmentExamMore(TestSegmentExam):
    def do_test_case_se2(self):
        """
        * Segementexam test with correlate layer-line option
        """
        self.feature_set.parameters['Pixel size in Angstrom']=self.tsr3d.pixelsize
        self.feature_set.parameters['Compute layer-line correlation option']=True
         
        super(TestSegmentExam, self).__init__(self.feature_set)
         
        self.add_up_power_spectra()
        
        
    def do_test_case_se3(self):
        """
        * Segementexam test with correlate layer-line option and reference power spectrum
        """
        self.add_up_power_spectra()
        
        shutil.copy('test_power.hdf', 'ref_power.hdf')
        self.feature_set.parameters['Reference power spectrum']=True
        super(TestSegmentExam, self).__init__(self.feature_set)
        
        self.add_up_power_spectra()
        os.remove('ref_power.hdf')


class TestSegmentExamEndToEnd(TestSegmentExam):
    def do_end_to_end_test_se_inputfile(self):
        EndToEndTest().do_end_to_end_inputfile(self.feature_set)

    def do_end_to_end_test_se_prompt(self):
        EndToEndTest().do_end_to_end_prompt(self.feature_set)

class TestSegmentExamMpi(TestSegmentExam):
    def do_end_to_end_test_sa_inputfile_with_MPI(self):
        self.feature_set.parameters['Compute layer-line correlation option']=True
        self.feature_set.parameters['MPI option'] = True
        self.feature_set.parameters['Number of CPUs']=2
        
        EndToEndTest().do_end_to_end_inputfile(self.feature_set)
        
def main():
    tse = TestSegmentExamMain()
    tse.setup()
    tse.do_test_case_se1()
        
if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Test module to check scanlinefit
"""
from EMAN2 import EMNumPy
from spring.csinfrastr.csproductivity import TestingDirectory
from spring.micprgs.scanlinefit import ScanLineFitPar, ScanLineFit
from spring.tests.csinfrastr.test_csreadinput import EndToEndTest
import numpy as np
import os

class TestScanLineFit(ScanLineFit):
    """
    Test class that tests work of scanlinefit
    """
    def setup(self):
        self.testingdir = TestingDirectory(os.path.splitext(os.path.basename(__file__))[0])
        self.testingdir.create()

        self.feature_set = ScanLineFitPar()
        self.prepare_line_micrograph()

        self.feature_set.parameters['Micrograph']='test_linemic.mrc'
        self.feature_set.parameters['Diagnostic plot']='test_scanlinefit.png'
        self.feature_set.parameters['Topleft coordinates of line']=((0,2000))
        self.feature_set.parameters['Bottomright coordinates of line']=((1498,201))
        self.feature_set.parameters['Width of integration in pixels']=15

        super(TestScanLineFit, self).__init__(self.feature_set)


    def prepare_line_micrograph(self):
        """
        Prepare a micrograph with defined line
        """
        self.linemic = np.zeros((2000,1500))
        self.icept = 2000
        self.slope = -1.2
        self.curvature = 1e-7
        self.pincushion = 1e-10

        polyvar = [self.pincushion, self.curvature, self.slope, self.icept]
        linex = np.arange(0,2000)
        self.line = np.polyval(polyvar, linex)

        # swap rows and col for EMData objects
        for row, col in enumerate(self.line):
            if 0 < row < 1500:
                self.linemic[np.rint(col)][row] += 1

        self.linemicimg = EMNumPy.numpy2em(np.copy(self.linemic))
        #self.linemicimg = mirror(self.linemicimg)
        self.linemicimg.write_image('test_linemic.mrc')
        """
        from EMAN2 import display
        display(self.linemicimg)
        #import matplotlib.pyplot as plt
        #plt.imshow(self.linemic)
        #plt.show()
        """
    def compare_fit(self):
        """
        * Function to compare fit with input data
        """
        assert round(self.pincushion, 9) == round(self.polyvar[0], 9)
        assert round(self.curvature, 5) == round(self.polyvar[1], 5)
        assert round (self.slope) == round(self.polyvar[2])
        assert round(self.icept, -3) == round(self.polyvar[3], -3)

    def teardown(self):
        os.remove('test_linemic.mrc')
        os.remove('test_scanlinefit.png')

        self.testingdir.remove()


class TestScanLineFitMain(TestScanLineFit):
    def do_test_case_lf1(self):
        self.perform_scanlinefit()
        self.compare_fit()


class TestScanLineFitMore(TestScanLineFit):
    def do_end_to_end_test_lf_inputfile(self):
        EndToEndTest().do_end_to_end_inputfile(self.feature_set)

    def do_end_to_end_test_lf_prompt(self):
        EndToEndTest().do_end_to_end_prompt(self.feature_set)

def main():
    tlf = TestScanLineFitMain()
    tlf.setup()
    tlf.do_test_case_lf1()
        
if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Test module to check new alignment module
"""
from EMAN2 import EMData
from fundamentals import rot_shift2D
from glob import glob
from multiprocessing import cpu_count
from spring.csinfrastr.csproductivity import TestingDirectory
from spring.segment2d.segmentalign2d import SegmentAlign2d, SegmentAlign2dPar
from spring.tests.csinfrastr.test_csreadinput import EndToEndTest
from spring.tests.segment2d.test_segmentclass import HelixClassPreparation
import os
import sys


class TestSegmentAlign2dPreparation(HelixClassPreparation):
    def apply_in_plane_rotation_in_degree_steps(self):
        self.helix_segments = EMData.read_images('test_segments.hdf')
        self.applied_rotations = []
        for image_number_on_stack, helix in enumerate(self.helix_segments):
            applied_angle = (-len(self.helix_segments) / 2 + image_number_on_stack) * self.rotation_step
            self.applied_rotations.append(applied_angle)
            helix = rot_shift2D(helix, angle=applied_angle, sx=7.0, sy=-1.2)
#            trans_2d = Transform({'type':'2d', 'alpha':-applied_angle, 'tx':6.0, 'ty':-1.2})
#            helix.set_attr('xform.align2d', trans_2d)
            helix.write_image(self.helix_stack, image_number_on_stack)
        

    def setup_reference_stack(self):
        self.reference = EMData()
        self.reference.read_image('test_segments.hdf', 0)
        self.reference.write_image('test_reference.hdf')
        

    def setup_helix_or_particle_dimension(self):
        self.feature_set.parameters['Estimated helix width and height in Angstrom'] = ((self.width_of_helix_in_angstrom,
        int(1.3 * self.width_of_helix_in_angstrom)))

    def setup_of_common_non_mpi_parameters(self):
        self.width_of_helix_in_angstrom = 500
        self.pixelsize = 5.0
        self.helix_stack = 'test_helix_stack.hdf'
        self.reference_stack = 'test_reference.hdf'
        self.aligned_ref = 'test_avg_ali.hdf'
        self.helix_stack_with_applied_alignment = 'test_helix_stack_aligned.hdf'
        self.align_directory = 'align_directory'
        self.custom_filter_file = 'test_filter_file.dat'
        self.restrain_option = True
        self.delta_in_plane_rotation = 30.0
        self.rotation_step = 2
        self.prepare_helix_projection()
        self.setup_reference_stack()
        micrograph_dimension = self.helix_micrograph.get_xsize()
        self.prepare_custom_filter_file(self.custom_filter_file, micrograph_dimension)
        self.apply_in_plane_rotation_in_degree_steps()
        self.logfile, self.directory = EndToEndTest().define_logfile_and_directory(self.feature_set.progname)
        self.feature_set.logfile = self.logfile
        self.feature_set.parameters['Image input stack'] = self.helix_stack
        self.feature_set.parameters['Image output stack'] = self.helix_stack_with_applied_alignment
        self.feature_set.parameters['Number of iterations'] = 2
        self.feature_set.parameters['Update references']=True
        self.feature_set.parameters['Aligned average stack'] = self.aligned_ref
        self.feature_set.parameters['Reference option'] = True
        self.feature_set.parameters['Image reference stack'] = self.reference_stack
        self.feature_set.parameters['Pixel size in Angstrom'] = self.pixelsize
        self.setup_helix_or_particle_dimension()
        
        self.feature_set.parameters['Internal binning factor']=1
        self.feature_set.parameters['Limit in-plane rotation']=self.restrain_option
        self.feature_set.parameters['Delta in-plane rotation angle'] = self.delta_in_plane_rotation
        self.feature_set.parameters['X and Y translation range in Angstrom'] = (100, 10)
        self.feature_set.parameters['Local refinement'] = True
        self.feature_set.parameters['Absolute X and Y translation limit in Angstrom']= (100, 10)
        
        self.feature_set.parameters['High-pass filter option'] = True
        self.feature_set.parameters['Low-pass filter option'] = True
        self.feature_set.parameters['High and low-pass filter cutoffs in 1/Angstrom'] = (0.01, 0.06)
        self.feature_set.parameters['Custom filter option'] = True
        self.feature_set.parameters['Custom-built filter file'] = self.custom_filter_file
        self.feature_set.parameters['Automatic filter option'] = True
        self.feature_set.parameters['B-Factor']=-2000
        self.feature_set.parameters['Temporary directory']=os.path.abspath(os.curdir)

    def teardown_segmentalign2d(self):
        pass
        os.remove('test_segments.hdf')
        os.remove(self.helix_stack)
        os.remove('test_reference.hdf')
        os.remove(self.custom_filter_file)
        os.remove(self.helix_stack_with_applied_alignment)
        if self.update_references:
            os.remove(self.aligned_ref)
            os.remove('{0}_var{1}'.format(os.path.splitext(self.aligned_ref)[0], os.path.splitext(self.aligned_ref)[-1]))
        
            add_files = glob('aqm*hdf') + glob('drm*dat') 
            for each_file in add_files:
                os.remove(each_file)

class TestSegmentAlign2d(TestSegmentAlign2dPreparation, SegmentAlign2d, EndToEndTest):

    def setup(self):
        self.testingdir = TestingDirectory(os.path.splitext(os.path.basename(__file__))[0])
        self.testingdir.create()

        self.feature_set = SegmentAlign2dPar()
        
        self.setup_of_common_non_mpi_parameters()
        self.feature_set.parameters['MPI option'] = False
        
        super(TestSegmentAlign2d, self).__init__(self.feature_set)
        

    def check_rotation(self, helix_stack, applied_rotations, delta_in_plane_rotation, rotation_step,
    restrain_inplane_rotation):
        helix_stack_aligned = EMData.read_images(helix_stack)
        for image_number_on_stack, helix in enumerate(helix_stack_aligned):
            alignment = helix.get_attr('xform.align2d')
            rotation_angle = alignment.get_params('2D')['alpha']
            translation_x = alignment.get_params('2D')['tx']
            translation_y = alignment.get_params('2D')['ty']
            
            applied_rotation_angle = -applied_rotations[image_number_on_stack]
            
            sys.stderr.write('\n{0}={1}; {2}, {3}'.format(rotation_angle %180, applied_rotation_angle %180,
            translation_x, translation_y))
            
            assert 360 != float(applied_rotation_angle) %360 


    def teardown(self):
        self.teardown_segmentalign2d()
 
        self.testingdir.remove()


class TestSegmentAlign2dMain(TestSegmentAlign2d):
    def do_test_case_sa1(self):

        self.perform_segmentalign2d()

        self.check_rotation(self.helix_stack, self.applied_rotations, self.delta_in_plane_rotation, self.rotation_step,
        self.restrain_option)
        
class TestSegmentAlign2dMore(TestSegmentAlign2d):
    def do_test_case_sa2(self):
        """
        * Single-CPU segmentalign test without reference and local refinement option
        """
        self.restrain_option=False
        self.feature_set.parameters['Limit in-plane rotation']=self.restrain_option
        self.feature_set.parameters['Reference option']= False
        self.feature_set.parameters['Local refinement'] = True
         
        super(TestSegmentAlign2d, self).__init__(self.feature_set)
         
        self.perform_segmentalign2d()
 
        self.check_rotation(self.helix_stack, self.applied_rotations, self.delta_in_plane_rotation, self.rotation_step,
        self.restrain_option)
         

    def do_test_case_sa3(self):
        """
        * Single-CPU segmentalign2d test without update option
        """
        self.feature_set.parameters['Update references']=False
        
        super(TestSegmentAlign2d, self).__init__(self.feature_set)
        
        self.perform_segmentalign2d()


class TestSegmentAlign2dEndToEnd(TestSegmentAlign2d):
    def do_end_to_end_test_sa_inputfile(self):
        EndToEndTest().do_end_to_end_inputfile(self.feature_set)
 
    def do_end_to_end_test_sa_prompt(self):
        EndToEndTest().do_end_to_end_prompt(self.feature_set)
        
class TestSegmentAlign2dMpi(TestSegmentAlign2d):
    def do_end_to_end_test_sa_inputfile_with_MPI(self):
        """
        * Standard Multi-CPU segmentalign2d run
        """
        self.feature_set.parameters['MPI option'] = True
        self.feature_set.parameters['Number of CPUs']=min(8, cpu_count())
        
        EndToEndTest().do_end_to_end_inputfile(self.feature_set)


    def do_end_to_end_test_sa_inputfile_with_MPI_2(self):
        """
        * Multi-CPU segmentalign2d run without reference (use average as starting reference)
        """
        self.feature_set.parameters['MPI option'] = True
        self.feature_set.parameters['Reference option']= False
        self.feature_set.parameters['Local refinement'] = True
        self.feature_set.parameters['Number of CPUs']=min(8, cpu_count())
        
        EndToEndTest().do_end_to_end_inputfile(self.feature_set)
        
        
    def do_end_to_end_test_sa_inputfile_with_MPI_3(self):
        """
        * Standard Multi-CPU segmentalign2d run
        """
        self.update_references = False
        self.feature_set.parameters['Update references']=self.update_references
        self.feature_set.parameters['MPI option'] = True
        self.feature_set.parameters['Number of CPUs']=min(8, cpu_count())
        
        EndToEndTest().do_end_to_end_inputfile(self.feature_set)


def main():
    talign = TestSegmentAlign2dMain()
    talign.setup()
    talign.do_test_case_sa1()
        
if __name__ == '__main__':
    main()



# Author: Carsten Sachse 21-Sep-2010
# Copyright: EMBL (2010 - 2018)
# License: see license.txt for details
"""
* Several classes to read input parameters from user: DefaultInput, PromptUser, ReadInputFile, ReadCommandLine \
(adapted from Hans-Peter Langtangen 'A Primer on Scientific Programming with Python')
"""
from glob import glob
from spring.csinfrastr.csfeatures import Features
from spring.csinfrastr.cslogger import Logger, FileCompleter, InputCompleter
from spring.csinfrastr.csproductivity import Errors
import os
import readline
import sys
import warnings

warnings.filterwarnings('ignore', module='matplotlib')
warnings.filterwarnings('ignore', module='sqlalchemy')

class CommandLineOptions:
    """
    * Class that defines common command line options
    """
    def define_options(self, progname, package, version_num, infile, outfile, proginfo):
        from argparse import ArgumentParser, SUPPRESS
        parser = ArgumentParser(prog=progname, description=proginfo) 

        parser.add_argument('input_output', nargs='*',
            help='Input and output files')
        parser.add_argument('--g', '--GUI', action='store_true', default=False, 
            help='GUI option: read input parameters from GUI', )
        parser.add_argument('--p', '--promptuser', action='store_true', default=False,
            help='Prompt user option: read input parameters from prompt', )
        parser.add_argument('--f', '--parameterfile', dest='filename',  
            help='File option: read input parameters from FILENAME')
        parser.add_argument('--c', '--cmd', action='store_true', default=False,
            help='Command line parameter option: read only boolean input parameters from command line and all other ' + \
            'parameters will be assigned from other sources', )
        parser.add_argument('--l', '--logfile', dest='logfilename', 
            help='Output logfile name as specified') 
        parser.add_argument('--d', '--directory', dest='directory_name', 
            help='Output directory name as specified') 

        #Print rst: to generate localized rst file for sphinx documentation
        parser.add_argument('--print_rst', action='store_true', help=SUPPRESS )
        
        parser.add_argument('--version', action='version', version='{0} from package {1}-{2}'.format(progname,
        package.title(), version_num))
        
        return parser

    def add_cmdline_options_from_parameter_dictionary(self):
        """
        * Function to add parameter-defined options to become command line options
        """
        parameters = [each_key for each_key in self.levels if self.levels[each_key] in ['beginner', 'intermediate', 'expert']]
        for name in parameters:
            help_string = self.hints[name] + ' (default: %(default)s)'
            if type(self.parameters[name]) is bool:
                try:
                    self.feature_set.parser.add_argument('--{0}'.format('_'.join(name.split()).lower()), 
                    '--{0}'.format(name[0:3].lower()), action='store_true', default=False, help=help_string)
                except:
                    self.feature_set.parser.add_argument('--{0}'.format('_'.join(name.lower().split())), 
                    action='store_true', default=False, help=help_string)
#             elif type(self.parameters[name]) is not bool:
#                 try:
#                     self.feature_set.parser.add_argument('--{0}'.format('_'.join(name.split()).lower()), 
#                     '--{0}'.format(name[0:3].lower()), dest='{0}'.format(name), 
#                     default=self.parameters[name], type=type(self.parameters[name]), help=help_string)
#                 except:
#                     self.feature_set.parser.add_argument('--{0}'.format('_'.join(name.lower().split())),
#                     dest='{0}'.format(name), default=self.parameters[name], type=type(self.parameters[name]),
#                     help=help_string)

    def override_cmdlineopts(self):
        # override boolean parameters from command line
        for name in self.parameters:
            if type(self.parameters[name]) is bool:
                if self.options.__dict__['_'.join(name.split()).lower()] is True:
                    self.parameters[name] = True
                else:
                    self.parameters[name] = False

        
class OptHandler(Features, CommandLineOptions):
    """
    * Class to handle the standard options of all cspackages

    #. Input: parser (contains all information about program options, parameter \
    dictionary, help dictionary, range dictionary)
    """

    def __init__(self, parset):
        self.feature_set = parset
        self.feature_set.default_params = parset.parameters.copy()

        self.parameters = parset.parameters
        self.levels = parset.level
        self.hints = parset.hints
        self.properties = parset.properties
        self.program_states = parset.program_states
        self.proginfo = parset.proginfo
        self.progname = parset.progname
        self.package = self.feature_set.eggmeta['Package']
        self.version = self.feature_set.eggmeta['Version']

        self.feature_set.parser = self.define_options(self.progname, self.feature_set.eggmeta['Package'], 
                                              self.feature_set.eggmeta['Version'], list(self.parameters.values())[0], 
                                              list(self.parameters.values())[1], parset.proginfo)
        self.add_cmdline_options_from_parameter_dictionary()

        self.evaluate_given_cmdline_option()
        self.make_and_change_to_new_working_directory()
        self.logfile = Logger().startlog(self.feature_set, debug=True)

    def evaluate_given_cmdline_option(self):
        """
        * Function to evaluate command line options and feed parameters in \
        program via chosen route 
        """
        self.options = self.feature_set.parser.parse_args()
        args = self.options.input_output

        if self.options.print_rst is True:
            from .csdocu import RecordRst
            RecordRst(self.feature_set)
        ######### Option handling
        # option 1: Command line
        if len(args) == 2:
            inp = ReadCommandLine(self.feature_set)
            self.options.c = True

        # option 2: read p from file
        if self.options.c is True:
            inp = DefaultInput(self.feature_set)
            self.override_cmdlineopts()
        elif self.options.filename is not None:
            self.feature_set.filename = self.options.filename
            inp = ReadInputFile(self.feature_set)
        # option 3: get p from prompt
        elif self.options.p is True: 
            inp = PromptUser(self.feature_set)
        # option 4: get p from GUI
        elif self.options.g is True: 
            self.startGui()
        else:
            self.feature_set.parser.print_help()
            sys.exit()

        self.feature_set.parameters = inp.parameters


    def make_and_change_to_new_working_directory(self):
        """
        * Function to make new working directory
        """
        from spring.csinfrastr.cslogger import NameGenerator
        
        self.change_filenames_to_absolute_paths()
                    
        if self.options.directory_name is None:
            new_directory_name = NameGenerator().make_directory(self.feature_set.progname)
        else:
            new_directory_name = self.options.directory_name
            
        if new_directory_name != os.path.curdir:
            try:
                os.mkdir(new_directory_name)
            except:
                error_message = 'Specified directory exists and cannot be overwritten. Please use different one.'
                raise IOError(error_message)
        
            os.chdir(new_directory_name)

    def change_filenames_to_absolute_paths(self):
        for self.name in self.parameters:
            if type(self.parameters[self.name]) is str:
                if self.properties[self.name].__str__().startswith('file'):
                    if self.properties[self.name].ftype in ['getFile', 'getFiles', 'getDir']:
                        if self.name in self.feature_set.relatives and \
                            self.parameters[self.feature_set.relatives[self.name]] is False:
                            pass
                        else:
                            if os.path.abspath(self.parameters[self.name]) == os.path.abspath(os.curdir):
                                self.parameters[self.name]=os.path.abspath(os.curdir)
                            elif not self.parameters[self.name].startswith(os.sep):
                                self.parameters[self.name]=os.path.join(os.path.abspath(os.curdir),
                                self.parameters[self.name])
                    elif self.properties[self.name].ftype in ['saveFile']:
                        if self.parameters[self.name].startswith(os.sep):
                            self.parameters[self.name]=os.path.basename(self.parameters[self.name])

    def startGui(self):
        """
        * Function to launch GUI
        """
        from PyQt4.QtGui import QApplication
        from PyQt4.QtCore import SIGNAL, SLOT
        from spring.csinfrastr.csgui import GUI
        
        app = QApplication(sys.argv)
        app.connect(app, SIGNAL('lastWindowClosed()') , app , SLOT('quit()'))
        self.window = GUI(self.feature_set)
        self.window.show()
        sys.exit(app.exec_())

class DefaultInput:
    """
    * Class to use default input parameters from program

    #. Input = default paramter dictionary
    #. Output = default parameter dictionary
    """
    def __init__(self, parset):
        print(parset.proginfo)
        print('Default input paramters demanded')
        self.parameters = parset.parameters


class PromptUser(Errors):
    """
    * Class to receive input parameters from prompt
    if no input is given default parameters are taken

    #. Input = default paramter dictionary
    #. Output = updated parameter dictionary
    """
    def __init__(self, parset):
        self.feature_set = parset
        self.parameters = parset.parameters
        self.levels = parset.level
        self.hints = parset.hints
        self.properties = parset.properties
        self.relatives = parset.relatives
        print(parset.proginfo)
        print('Please enter the input parameters at the prompt (ENTER for default, ' + \
              'Hit Tab for suggestions and autocomplete):')
        
        readline.set_completer_delims(' \t\n;')
        readline.parse_and_bind("tab: complete")
        
        self.prompt_user()

    def check_numbers_and_boolean_input(self, entered):
        if type(self.parameters[self.name]) is tuple:
            super(PromptUser, self).check_valerr(eval(entered)[0])
            super(PromptUser, self).check_valerr(eval(entered)[1])
            if type(self.parameters[self.name][0]) is float:
                self.parameters[self.name]=(float(eval(entered)[0]), float(eval(entered)[1]))
            elif type(self.parameters[self.name][0]) is int:
                self.parameters[self.name]=(int(eval(entered)[0]), int(eval(entered)[1]))
            else:
                self.parameters[self.name] = eval(entered)
        elif type(self.parameters[self.name]) is float:
            super(PromptUser, self).check_valerr(entered)
            self.parameters[self.name] = float(entered)
        elif type(self.parameters[self.name]) is int:
            super(PromptUser, self).check_valerr(entered)
            self.parameters[self.name] = int(entered)
        elif type(self.parameters[self.name]) is bool:
            if str(entered).title() in ['True', 'Yes', 'Y']:
                self.parameters[self.name] = True
            elif str(entered).title() in ['False', 'No', 'N']:
                self.parameters[self.name] = False
            else:
                error_message = 'Input not understood: True/False or Yes/No.'
                sys.stderr.write(error_message)
                raise TypeError(error_message)

    def check_prompt_against_expected_strings(self, entered):
        if type(self.parameters[self.name]) is str:
            entered = entered.strip().strip('\n')
            if self.properties[self.name].__str__().startswith('file'):
                if entered.find(',') >= 0:
                    file_list = entered.split(',')
                elif entered.find('*') >= 0 or entered.find('?') >= 0:
                    file_list = glob(entered)
                    file_list.sort()
                    if file_list == []:
                        file_list = [entered]
                else:
                    file_list = [entered]

                for each_file in file_list:
                    if self.properties[self.name].ftype in ['getFile', 'getFiles', 'saveFile']:
                        super(PromptUser, self).check_fileformat(each_file)
                        if self.properties[self.name].ftype not in ['saveFile']:
                            super(PromptUser, self).check_filerr(each_file)
                    elif self.properties[self.name].ftype in ['getDir']:
                        super(PromptUser, self).check_filerr(each_file)
                
                self.parameters[self.name] = entered
            elif self.properties[self.name].__str__().startswith('choice'):
                if str(entered).lower() in self.properties[self.name].choices:
                    self.parameters[self.name] = str(entered).lower()
                else:
                    error_message = 'Input not understood: {0}.'.format(self.properties[self.name].choices.__str__())
                    sys.stderr.write(error_message)
                    raise TypeError(error_message)

    def setup_tab_autocompleter(self):
        if type(self.parameters[self.name]) is bool:
            self.comp = InputCompleter(['true', 'false', 'True', 'False'])
        elif type(self.parameters[self.name]) is str:
            if self.properties[self.name].__str__().startswith('choice'):
                self.comp = InputCompleter(self.properties[self.name].choices)
            elif self.properties[self.name].__str__().startswith('file'):
                if self.properties[self.name].ftype in ['getFile', 'getFiles', 'saveFile', 'getDir']:
                    self.comp = FileCompleter()
                    readline.set_completer(self.comp.complete)
                else:
                    self.comp = InputCompleter([str(self.parameters[self.name])])
        else:
            self.comp = InputCompleter([str(self.parameters[self.name])])
        readline.set_completer(self.comp.complete)
        entered = raw_input('Please enter {0}: {1}= \n>>> '.format(self.name, self.parameters[self.name]))
        entered = str(entered).strip()
        return entered

    def prompt_user(self):
        """
        * Function to prompt user on parameter input, performs parameter \
        checking based on type of input and checks whether entered values are \
        in expected range

        """
        parameters = [each_key for each_key in self.levels if self.levels[each_key] in ['beginner', 'intermediate', 'expert']]
        for self.name in parameters:
            while True:
                try:
                    if type(self.parameters[self.name]) is not tuple and self.relatives.has_key(self.name):
                        if self.parameters[self.relatives[self.name]] is False:
                            break
                    elif type(self.parameters[self.name]) is tuple and self.relatives.has_key(self.name):
                        if self.parameters[self.relatives[self.name][0]] is False and \
                        self.parameters[self.relatives[self.name][1]] is False:
                            break
                    print('\n' + self.hints[self.name])
                    entered = self.setup_tab_autocompleter()
                    if entered.lower() in ['quit', 'end', 'quit()', 'end()']:
                        raise EOFError
                    elif entered is not '':
                        self.check_numbers_and_boolean_input(entered)
                        self.check_prompt_against_expected_strings(entered)
                    else:
                        break
                    break
                except EOFError:
                    raise SystemExit('\n{0} terminated.'.format(self.feature_set.progname))
                except:
                    print ' Hit Tab to get suggestions and auto-complete.'
                    pass

        return self.parameters

class ReadInputFile:
    """
    Class to receive input parameters from an input file
    input = default parameter dictionary, filename
    output = updated parameter dictionary
    """
    def __init__(self, parset):
        self.parameters = parset.parameters
        self.properties = parset.properties
        print(parset.proginfo)
        print('Input parameters are read from {0}'.format(parset.filename))
        self.read_file(parset.filename)
    
    def clear_parameters(self, parameters):
        for name in parameters:
            parameters[name]=None
        
        return parameters
    
    def evaluate_value_type(self, value, teipp):
        try:
            teipp(value)
        except ValueError:
            error_message = ('\'{0}\' received \'{1}\'\nPlease enter {2} value from {3} to {4}'.format(self.name, value, 
                                        teipp, self.properties[self.name].minimum, self.properties[self.name].maximum))
            raise ValueError(error_message)


    def enter_received_value_if_it_is_tuple_of_numbers(self, received_value, type_of_value):
        if type_of_value is tuple:
            received_value = received_value.strip('(')
            received_value = received_value.strip(')')
            if ',' in received_value:
                valentries = received_value.split(',')
            if ';' in received_value:
                valentries = received_value.split(';')
            self.evaluate_value_type(valentries[0], float)
            self.evaluate_value_type(valentries[-1], float)
            if type(self.default_params[self.name][0]) is float:
                self.parameters[self.name]=(float(eval(received_value)[0]), float(eval(received_value)[1]))
            elif type(self.default_params[self.name][0]) is int:
                self.parameters[self.name]=(int(eval(received_value)[0]), int(eval(received_value)[1]))
            else:
                self.parameters[self.name] = eval(received_value)
            
        return received_value


    def enter_received_value_if_it_is_boolean_option(self, received_value, type_of_value):
        if type_of_value is bool:
            if received_value.strip() in ['True', 'Yes', 'Y', 'yes', 'y']:
                self.parameters[self.name] = True
            elif received_value.strip() in ['False', 'No', 'N', 'no', 'n']:
                self.parameters[self.name] = False

    def enter_received_value_if_it_is_number(self, received_value, type_of_value):
        if type_of_value is float:
            self.evaluate_value_type(received_value, float)
            self.parameters[self.name] = float(received_value)
        elif type_of_value is int:
            self.evaluate_value_type(received_value, float)
            self.parameters[self.name] = int(float(received_value))


    def enter_received_value_if_it_is_file_or_available_choice(self, received_value, type_of_value):
        if type_of_value is str:
            if self.properties[self.name].__str__().startswith('choice'):
                if received_value.lower() not in self.properties[self.name].choices:
                    error_message = 'Choice \'{0}\' not referenced in default parameter dictionary. Choose between {1}'\
                    .format(received_value, self.properties[self.name].choices.__str__())
                    raise ValueError(error_message)
                else:
                    self.parameters[self.name] = received_value.lower()
            else:
                self.parameters[self.name] = received_value


    def read_file(self, filename):
        infile = open( filename, 'r')  
         
        self.default_params = self.parameters.copy()
        self.parameters = self.clear_parameters(self.parameters)
        for line in infile:
            if '=' in line:
                name, received_value = line.split('=')
                self.name = name.strip()
                received_value = received_value.strip()
                if self.name in self.parameters:
                    if received_value == 'None':
                        pass
                    else:
                        type_of_value = type(self.default_params[self.name])
                        self.enter_received_value_if_it_is_tuple_of_numbers(received_value, type_of_value)
                        self.enter_received_value_if_it_is_number(received_value, type_of_value)
                        self.enter_received_value_if_it_is_boolean_option(received_value, type_of_value)
                        self.enter_received_value_if_it_is_file_or_available_choice(received_value, type_of_value)
                else:
                    print('Parameter \'{0}\' is of no use for the program because not '.format(self.name)) + \
                          'referenced in default parameter dictionary'
        
        return self.parameters

class ReadCommandLine:
    """
    * Class to receive input parameters from the command line

    #. Input = default parameter dictionary
    #. Output = updated parameter dictionary
    """
    def __init__(self, parset):
        self.parameters = parset.parameters
        self.parser = parset.parser
        options = self.parser.parse_args()
        self.args = options.input_output
        print(parset.proginfo)
        print('Input and output files detected from command line')
        self._read_command_line()

    def _read_command_line(self):
        self.parameters[list(self.parameters.keys())[0]] = self.args[0]
        self.parameters[list(self.parameters.keys())[1]] = self.args[1]
        return self.parameters




# Author: Carsten Sachse 08-Jun-2011
# Copyright: EMBL (2010 - 2018)
# License: see license.txt for details
from EMAN2 import EMData, Util, Transform
from collections import namedtuple
from spring.csinfrastr.csdatabase import SpringDataBase, base, SegmentTable, RefinementCycleSegmentTable
from spring.segment2d.segment import Segment
from spring.segment2d.segmentalign2d import SegmentAlign2d
from spring.segment2d.segmentselect import SegmentSelect
from spring.segment3d.refine.sr3d_project import SegmentRefine3dProjectionLayerLineFilter
from spring.segment3d.segclassreconstruct import SegClassReconstruct
from sqlalchemy.sql.expression import asc
from tabulate import tabulate
from utilities import compose_transform2
import numpy as np
import os
import sys


class SegmentRefine3dAlignMatch(SegmentRefine3dProjectionLayerLineFilter):
    def prepare_segment_for_alignment(self, masked_segment_stack, previous_params, segment_size, each_info,
    limit_search_ranges=True):
        segment = EMData()
        segment.read_image(masked_segment_stack, previous_params.local_id)

        if limit_search_ranges:
            x_distance, y_distance = SegClassReconstruct().compute_distances_to_helical_axis(previous_params.shift_x,
            previous_params.shift_y, previous_params.inplane_angle)
                 
            x_range = float(min(max(0, self.x_limit_A / each_info.pixelsize - abs(x_distance)), each_info.x_range))
            y_range = float(min(max(0, self.y_limit_A / each_info.pixelsize - abs(y_distance)), each_info.y_range))
        else:
            x_range = each_info.x_range
            y_range = each_info.y_range

        segment = Segment().shift_and_rotate_image(segment, previous_params.inplane_angle, previous_params.shift_x,
        previous_params.shift_y)
        
        center_x = center_y = segment_size / 2 + 1
        
        return segment, center_x, center_y, x_range, y_range
    

    def convert_Euler_angles_to_mirrored_sparx(self, phi, theta, psi):
        """
        >>> from spring.segment3d.refine.sr3d_main import SegmentRefine3d
        >>> s = SegmentRefine3d()
        >>> phi, theta, psi = s.convert_Euler_angles_from_mirrored_sparx(90, 90, 270)
        >>> s.convert_Euler_angles_to_mirrored_sparx(phi, theta, psi)
        (90.0, 90.0, 270.0)
        """
        mirrored_phi = (phi - 540.0) % 360
        mirrored_theta = (180.0 - theta) % 360
        mirrored_psi = (540.0 - psi) % 360
        
        return mirrored_phi, mirrored_theta, mirrored_psi
    

    def convert_Euler_angles_from_mirrored_sparx(self, phi, theta, psi):
        """
        >>> from spring.segment3d.refine.sr3d_main import SegmentRefine3d
        >>> s = SegmentRefine3d()
        >>> s.convert_Euler_angles_from_mirrored_sparx(90, 90, 270)
        (270.0, 90.0, 270.0)
        >>> s.convert_Euler_angles_from_mirrored_sparx(np.linspace(0, 360, 2), np.linspace(82, 98, 2), 270 * np.ones(2))
        (array([ 180.,  180.]), array([ 98.,  82.]), array([ 270.,  270.]))
        """
        mirrored_phi = (phi + 540.0) % 360
        mirrored_theta = (180.0 - theta) % 360
        mirrored_psi = (540.0 - psi) % 360
        
        return mirrored_phi, mirrored_theta, mirrored_psi
    

    def combine_determined_parameters_to_final_five(self, previous_params, reference_rings, projection_parameters, ang,
    sxs, s_ys, mirror, matched_reference_id, peak):
        """
        Function to convert (in-plane rotation and shifts) to (shifts and rotation)
        if mirror parameters have been found, adjust Euler angles    
        """
        peak = abs(peak)
        matched_reference_id = int(matched_reference_id)
        model = projection_parameters[matched_reference_id].model_id
        if matched_reference_id > -1:
            angb, sxb, syb, ct = compose_transform2(0.0, sxs, s_ys, 1, -ang, 0.0, 0.0, 1)
            inplane_angle = (ang) % 360
            if mirror:
                phi, theta, psi = \
                self.convert_Euler_angles_from_mirrored_sparx(projection_parameters[matched_reference_id].phi,
                projection_parameters[matched_reference_id].theta, projection_parameters[matched_reference_id].psi +
                angb)
            else:
                phi = projection_parameters[matched_reference_id].phi
                theta = projection_parameters[matched_reference_id].theta
                psi = (projection_parameters[matched_reference_id].psi + angb) % 360
            s2x = (sxb + previous_params.shift_x)
            s2y = (syb + previous_params.shift_y)
        else:
            phi = theta = psi = s2x = s2y = t1 = 0.0
            peak = -1.0e23

        return model, phi, theta, psi, s2x, s2y, inplane_angle, peak


    def find_inplane_to_match(self, phiA,thetaA,phiB,thetaB,psiA=0,psiB=0):
        """Find the z rotation such that
            ZA  RA is as close as possible to RB
                this maximizes trace of ( RB^T ZA RA) = trace(ZA RA RB^T)
        """
    
        RA   = Transform({'type': 'spider', 'phi': phiA, 'theta': thetaA, 'psi': psiA})
        RB   = Transform({'type': 'spider', 'phi': phiB, 'theta': thetaB, 'psi': psiB})
        RBT  = RB.transpose()
        RABT = RA * RBT
    
        RABTeuler = RABT.get_rotation('spider')
        RABTphi   = RABTeuler['phi']
        RABTtheta = RABTeuler['theta']
        RABTpsi   = RABTeuler['psi']
    
        return (-RABTpsi-RABTphi),RABTtheta
    

    def filter_angle_ids_according_to_search_restraint(self, phis, phi_ids, each_phi, phi_restraint,
    max_angle=360.0):
        """
        >>> from spring.segment3d.refine.sr3d_main import SegmentRefine3d
        >>> s = SegmentRefine3d()
        >>> s.filter_angle_ids_according_to_search_restraint(np.arange(0, \
360.0, 10), np.arange(36), 55.0, 80.0)
        array([  0.,   1.,   2.,   3.,   4.,   5.,   6.,   7.,   8.,   9.,  10.,
                11.,  12.,  13.,  34.,  35.])
                
        >>> s.filter_angle_ids_according_to_search_restraint(np.arange(0, \
360.0, 10), np.arange(36), 335.0, 80.0)
        array([  0.,   1.,   2.,   3.,   4.,   5.,  26.,  27.,  28.,  29.,  30.,
                31.,  32.,  33.,  34.,  35.])
                
        >>> s.filter_angle_ids_according_to_search_restraint(np.arange(0, \
360.0, 10), np.arange(36), 135.0, 80.0)
        array([  6.,   7.,   8.,   9.,  10.,  11.,  12.,  13.,  14.,  15.,  16.,
                17.,  18.,  19.,  20.,  21.])
                
        >>> s.filter_angle_ids_according_to_search_restraint(np.arange(82, \
99), np.arange(34), 82.0, 3.0)
        array([ 0.,  1.,  2.])
        """
        lower_bound = (each_phi - phi_restraint) 
        higher_bound = (each_phi + phi_restraint) 
        filtered_ids = phi_ids[(lower_bound < phis) & (phis < higher_bound)]
        
        add_low_filtered = np.array([])
        add_high_filtered = np.array([])
            
        if lower_bound != (lower_bound) % max_angle:
            add_low_filtered = phi_ids[((lower_bound) % max_angle < phis)] 
        if higher_bound != (higher_bound) % max_angle:
            add_high_filtered = phi_ids[((higher_bound) % max_angle > phis)] 
        
        filtered_ids = np.append(filtered_ids, add_low_filtered)
        filtered_ids = np.append(filtered_ids, add_high_filtered)
        filtered_ids = np.sort(np.unique(filtered_ids))
        
        return filtered_ids
        

    def compute_distance(self, this_phi, this_theta, phis, thetas):
        distances = np.sqrt((phis - this_phi) ** 2 + (thetas - this_theta) ** 2)
        
        return distances 
        

    def get_closest_projection_from_projection_list(self, this_phi, this_theta, phis, thetas, ids):
        """
        >>> from spring.segment3d.refine.sr3d_main import SegmentRefine3d
        >>> s = SegmentRefine3d()
        >>> phis = np.arange(10, 90, 10)
        >>> thetas = np.array(8 * [90.0])
        >>> ids = np.arange(len(phis))
        >>> s.get_closest_projection_from_projection_list(350, 90, phis, thetas, ids)
        (0, 10, 90.0)
        
        >>> phis = np.array(8 * np.arange(10, 90, 10).tolist())
        >>> thetas = np.array([each_theta for each_theta in np.arange(82, 98, 2) for each_time in range(8)])
        >>> ids = np.arange(len(phis))
        >>> s.get_closest_projection_from_projection_list(350, 98, phis, thetas, ids)
        (56, 10, 96)
        """
        distances = self.compute_distance(this_phi, this_theta, phis, thetas)
        distances = np.append(distances, self.compute_distance(this_phi + 360.0, this_theta, phis, thetas))
        distances = np.append(distances, self.compute_distance(this_phi - 360.0, this_theta, phis, thetas))
        
        ids = np.append(ids, np.append(ids, ids))
        closest_index = np.argmin(distances)
        closest_id = ids[closest_index]
        closest_phi = phis[closest_id]
        closest_theta = thetas[closest_id]
        
        return closest_id, closest_phi, closest_theta
    
        
    def select_reference_rings_according_to_angular_restraints(self, reference_rings, each_info, projection_parameters,
    each_proj_parameter):
        """
        >>> from spring.segment3d.refine.sr3d_main import SegmentRefine3d
        >>> s = SegmentRefine3d()
        >>> ref_rings = ['first_ring', 'second_ring']
        >>> restraint_tuple = s.get_restraint_tuple()
        >>> info = restraint_tuple(20, 20, 5, 5, 5)
        >>> proj_tuple = s.get_prj_params_named_tuple()
        >>> proj_params = [proj_tuple(0, 0, 0, 90, 270, 0, 0), proj_tuple(1, 0, 350, 90, 270, 0, 0)]
        >>> last_param = proj_tuple(0, 0, 0, 90, 270, 0, 0)
        >>> s.select_reference_rings_according_to_angular_restraints(ref_rings, info, proj_params, last_param) #doctest: +NORMALIZE_WHITESPACE
        (['first_ring', 'second_ring'], 
        [prj_parameters(id=0, model_id=0, phi=0, theta=90, psi=270, shift_x=0, shift_y=0), 
        prj_parameters(id=1, model_id=0, phi=350, theta=90, psi=270, shift_x=0, shift_y=0)], 
        ('340.00 - 20.00', '90.00 - 90.00'))
        
        >>> proj_params = [proj_tuple(0, 0, 0, 60, 270, 0, 0), proj_tuple(1, 0, 0, 90, 270, 0, 0)]
        >>> s.select_reference_rings_according_to_angular_restraints(ref_rings, info, proj_params, last_param) #doctest: +NORMALIZE_WHITESPACE
        (['second_ring'], 
        [prj_parameters(id=1, model_id=0, phi=0, theta=90, psi=270, shift_x=0, shift_y=0)], 
        ('340.00 - 20.00', '70.00 - 110.00'))
        
        >>> proj_params = [proj_tuple(0, 0, 0, 60, 270, 0, 0), proj_tuple(1, 0, 330, 90, 270, 0, 0)]
        >>> s.select_reference_rings_according_to_angular_restraints(ref_rings, info, proj_params, last_param) #doctest: +NORMALIZE_WHITESPACE
        (['first_ring'], 
        [prj_parameters(id=0, model_id=0, phi=0, theta=60, psi=270, shift_x=0, shift_y=0)], 
        ('340.00 - 20.00', '70.00 - 110.00'))
        
        >>> proj_params = [proj_tuple(0, 1, 0, 90, 270, 0, 0), proj_tuple(1, 0, 350, 90, 270, 0, 0)]
        >>> s.select_reference_rings_according_to_angular_restraints(ref_rings, info, proj_params, last_param) #doctest: +NORMALIZE_WHITESPACE
        (['second_ring'], 
        [prj_parameters(id=1, model_id=0, phi=350, theta=90, psi=270, shift_x=0, shift_y=0)], 
        ('340.00 - 20.00', '90.00 - 90.00'))
        """
        
        if each_info.azimuthal_restraint == 180.0 and each_info.out_of_plane_restraint == 180.0:
            selected_reference_rings = reference_rings
            selected_projection_parameters = projection_parameters
            selected_log_info = (None, None)
        else:
            phis = np.array([each_prj.phi for each_prj in projection_parameters])# if each_proj_parameter.model_id == each_prj.model_id])
            thetas = np.array([each_prj.theta for each_prj in projection_parameters])# if each_proj_parameter.model_id == each_prj.model_id])
            psis = np.array([each_prj.psi for each_prj in projection_parameters])# if each_proj_parameter.model_id == each_prj.model_id])
            model_ids = np.array([each_prj.model_id for each_prj in projection_parameters])# if each_proj_parameter.model_id == each_prj.model_id])
            
            mirr_phis, mirr_thetas, mirr_psis = self.convert_Euler_angles_to_mirrored_sparx(phis, thetas, psis)
            prj_ids = np.append(np.arange(len(phis)), np.arange(len(mirr_phis)))
            phis = np.append(phis, mirr_phis)
            thetas = np.append(thetas, mirr_thetas)
            model_ids = np.append(model_ids, model_ids)
            
            filtered_phi_ids = self.filter_angle_ids_according_to_search_restraint(phis, prj_ids, 
            each_proj_parameter.phi, each_info.azimuthal_restraint)
            
            filtered_theta_ids = self.filter_angle_ids_according_to_search_restraint(thetas, prj_ids,
            each_proj_parameter.theta, each_info.out_of_plane_restraint)
            
            filtered_model_ids = prj_ids[model_ids == each_proj_parameter.model_id]
            
            filtered_prj_ids = np.intersect1d(filtered_phi_ids, np.intersect1d(filtered_theta_ids, filtered_model_ids))
            
            if reference_rings is not None:
                selected_reference_rings = [reference_rings[int(each_selected_prj)] for each_selected_prj in
                filtered_prj_ids]
            else:
                selected_reference_rings = None
            
            selected_projection_parameters = [projection_parameters[int(each_selected_prj)] for each_selected_prj in
            filtered_prj_ids]
        
            if selected_projection_parameters == []:
                filtered_phis = phis[model_ids == each_proj_parameter.model_id]
                filtered_thetas = thetas[model_ids == each_proj_parameter.model_id]
                closest_id, closest_phi, closest_theta = \
                self.get_closest_projection_from_projection_list(each_proj_parameter.phi, each_proj_parameter.theta,
                filtered_phis, filtered_thetas, filtered_model_ids)
                
                selected_projection_parameters = [projection_parameters[closest_id]]
                if reference_rings is not None:
                    selected_reference_rings = [reference_rings[closest_id]]
                
            selected_phi_str = '{0:.2f} - {1:.2f}'.format((each_proj_parameter.phi - each_info.azimuthal_restraint) % 360,
                                                   (each_proj_parameter.phi + each_info.azimuthal_restraint) % 360)
            
            selected_theta_str = '{0:.2f} - {1:.2f}'.\
            format(max(np.min(thetas), (each_proj_parameter.theta - each_info.out_of_plane_restraint) % 360),
            (min(np.max(thetas), (each_proj_parameter.theta + each_info.out_of_plane_restraint) % 360)))
            
            selected_log_info = (selected_phi_str, selected_theta_str)
            
        return selected_reference_rings, selected_projection_parameters, selected_log_info
        
        
    def match_each_image_against_all_projections_with_delta_in_plane_restraint(self, masked_segment_stack,
    previous_params, reference_rings, projection_parameters, each_info, step_x, delta_psi,
    polar_interpolation_parameters, segment_size, full_circle_mode='F'):
        
        segment, center_x, center_y, x_range, y_range = self.prepare_segment_for_alignment(masked_segment_stack,
        previous_params, segment_size, each_info)
        
        [ang, sxs, s_ys, mirror, matched_reference_id, peak] = Util.multiref_polar_ali_2d_delta(segment,
        reference_rings, x_range, y_range, step_x, full_circle_mode, polar_interpolation_parameters, 
        center_x, center_y, 0.0, delta_psi)
        
        model, phi, theta, psi, s2x, s2y, inplane_angle, peak = \
        self.combine_determined_parameters_to_final_five(previous_params, reference_rings, projection_parameters, 
        ang + previous_params.inplane_angle, sxs, s_ys, mirror, matched_reference_id, peak)
        
        return [model, phi, theta, psi, s2x, s2y, inplane_angle, peak, mirror]
    

    def refine_each_image_against_projections_including_delta_inplane_restraint(self, masked_segment_stack,
    previous_params, reference_rings, projection_parameters, each_info, step_x, step_angle, delta_psi,
    polar_interpolation_parameters, segment_size, full_circle_mode='F'):
        
        segment, center_x, center_y, x_range, y_range = self.prepare_segment_for_alignment(masked_segment_stack,
        previous_params, segment_size, each_info)
    
        transform_projection = Transform({'type':'spider','phi':previous_params.phi,'theta':previous_params.theta,'psi':
        (270.0)%360})
        
        segment.set_attr('xform.projection', transform_projection)
        
        [ang, sxs, s_ys, mirror, matched_reference_id, peak] = Util.multiref_polar_ali_2d_local(segment,
        reference_rings, x_range, y_range, float(step_x), step_angle, full_circle_mode,
        polar_interpolation_parameters, center_x, center_y)
        
        model, phi, theta, psi, s2x, s2y, inplane_angle, peak = \
        self.combine_determined_parameters_to_final_five(previous_params, reference_rings, projection_parameters, 
        ang + previous_params.inplane_angle, sxs, s_ys, mirror, matched_reference_id, peak)
        
        return [model, phi, theta, psi, s2x, s2y, inplane_angle, peak, mirror]
    
    
class SegmentRefine3dAlign(SegmentRefine3dAlignMatch):
    def set_euler_angles_to_reference_rings(self, reference_rings, reference_angles):

        for each_ref_id, each_ref_angle in enumerate(reference_angles):
            n1 = np.sin(np.deg2rad(each_ref_angle.theta)) * np.cos(np.deg2rad(each_ref_angle.phi))
            n2 = np.sin(np.deg2rad(each_ref_angle.theta)) * np.sin(np.deg2rad(each_ref_angle.phi))
            n3 = np.cos(np.deg2rad(each_ref_angle.theta))
            reference_rings[each_ref_id].set_attr_dict({'n1':n1, 'n2':n2, 'n3':n3})
            reference_rings[each_ref_id].set_attr('phi', each_ref_angle.phi)
            reference_rings[each_ref_id].set_attr('theta', each_ref_angle.theta)
            reference_rings[each_ref_id].set_attr('psi', each_ref_angle.psi)

        return reference_rings
    
    
    def prepare_reference_rings_from_projections(self, projection_stack, projection_parameters, alignment_size,
    max_range):
        first_ring = 1
        last_ring = max(int(alignment_size * 0.75 / 2), alignment_size / 2 - int(np.sqrt(2) * max_range))
        ring_step = 1
        
        polar_interpolation_parameters, ring_weights = SegmentAlign2d().prepare_empty_rings(first_ring, last_ring,
        ring_step)
        
        prj = EMData()
        reference_rings = []
        for each_prj in projection_parameters:
            prj.read_image(projection_stack, each_prj.id)
            
            cimage = SegmentAlign2d().generate_reference_rings_from_image(prj, polar_interpolation_parameters,
            ring_weights, alignment_size)
            
            reference_rings.append(cimage)

        reference_rings = self.set_euler_angles_to_reference_rings(reference_rings, projection_parameters)

        return reference_rings, polar_interpolation_parameters
    
    
    def make_named_tuple_of_orientation_parameters(self):
        orientation_parameters = namedtuple('orientation', 'stack_id local_id model_id phi theta psi shift_x ' + \
        'shift_y inplane_angle peak mirror rank_id')
        
        return orientation_parameters
    

    def get_masking_parameters(self, masking_parameters, mask_params, each_segment, pixelsize):
        if each_segment.lavg_inplane_angle is not None:
            inplane_angle = each_segment.lavg_inplane_angle
            shift_x_A = each_segment.lavg_helix_shift_x_A
        else:
            inplane_angle = each_segment.inplane_angle
            shift_x_A = each_segment.shift_x_A

        mask_params.append(masking_parameters(each_segment.stack_id, each_segment.stack_id, inplane_angle, shift_x_A /
        pixelsize, abs(each_segment.theta - 90.0)))
        
        return mask_params
    

    def make_named_tuple_of_masking_parameters(self):
        masking_parameters = namedtuple('mask', 'stack_id local_id lavg_inplane_angle lavg_helix_shift_x ' +
        'out_of_plane_tilt')
        
        return masking_parameters
    

    def compute_optimum_phi_and_index_closest_to_one_of_prj_phis(self, multiples, phis, prj_phis):
        """
        >>> from spring.segment3d.refine.sr3d_main import SegmentRefine3d
        >>> s = SegmentRefine3d()
        >>> mult = np.array([0, 1, 2]) 
        >>> phis = np.array([325, 97, 355])
        >>> prj_phis = np.array([0, 90, 180, 270])
        >>> s.compute_optimum_phi_and_index_closest_to_one_of_prj_phis(mult, phis, prj_phis)
        (355, 2)
        """
        ang_distances = [np.min(np.abs(phis - each_prj_phi)) for each_prj_phi in prj_phis]
        ang_distances += [np.min(np.abs(phis - each_prj_phi + 360)) for each_prj_phi in prj_phis]
        ang_distances += [np.min(np.abs(phis - each_prj_phi - 360)) for each_prj_phi in prj_phis]
        
#        min_ang_dist = np.min(ang_distances)
        
        closest_prj_indices = [np.argmin(np.abs(phis - each_prj_phi)) for each_prj_phi in prj_phis]
        closest_prj_indices += [np.argmin(np.abs(phis - each_prj_phi + 360)) for each_prj_phi in prj_phis]
        closest_prj_indices += [np.argmin(np.abs(phis - each_prj_phi - 360)) for each_prj_phi in prj_phis]
        closest_index = closest_prj_indices[np.argmin(ang_distances)]
        closest_phi = phis[closest_index]
        closest_multiple = multiples[closest_index]
        
        return closest_phi, closest_multiple
    
        
    def reduce_y_shifts_and_phis_to_half_stepsize(self, orientation_result, helix_y_shifts, helical_symmetry,
    projection_parameters, stepsize):
        """
        >>> from spring.segment3d.refine.sr3d_main import SegmentRefine3d
        >>> n_phi = namedtuple('orient', 'phi')
        >>> o_rslt = [n_phi(5), n_phi(87), n_phi(330), n_phi(350)]
        >>> y_shifts = np.array([1, 2, 4, 5])
        >>> prj_params = [[0], [90], [180], [270]]
        >>> s = SegmentRefine3d()
        >>> s.reduce_y_shifts_and_phis_to_half_stepsize(o_rslt, y_shifts, (1, 10), prj_params, 10) 
        (array([ 1.,  2., -2., -3.]), array([   5.,   87.,  270.,  270.]))
        >>> s.reduce_y_shifts_and_phis_to_half_stepsize(o_rslt, y_shifts, (0, 10), prj_params, 10) 
        (array([ 1.,  2.,  4.,  5.]), array([   5.,   87.,  330.,  350.]))
        """
        helical_rise, helical_rotation = helical_symmetry
        phis = np.array([each_result.phi for each_result in orientation_result])
        if helical_rise != 0:
            multiples = np.round(helix_y_shifts / helical_rise)
        else:
            multiples = np.zeros(len(helix_y_shifts))
        close_to_0_ys = helix_y_shifts - helical_rise * multiples
        close_to_0_phis = (phis - multiples * helical_rotation) %360
        prj_phis = np.unique([each_prj[0] for each_prj in projection_parameters])
        
        if helical_rise != 0:
            stepsize_multiple = np.round(stepsize / helical_rise)
            multiples_to_test = np.arange(stepsize_multiple) - stepsize_multiple / 2
        
            optimum_multiples = np.array([])
            for each_stack_id, each_multiple in enumerate(multiples):
                each_phi = close_to_0_phis[each_stack_id]
                phis_to_test = (each_phi + multiples_to_test * helical_rotation) %360
                
                optimum_phi, optimum_multiple = \
                self.compute_optimum_phi_and_index_closest_to_one_of_prj_phis(multiples_to_test, phis_to_test, prj_phis)
                
                optimum_multiples = np.append(optimum_multiples, optimum_multiple)
        else:
            optimum_multiples = 0
                
        reduced_y_shifts = close_to_0_ys + helical_rise * optimum_multiples
        reduced_phis = (close_to_0_phis + optimum_multiples * helical_rotation) %360

        return reduced_y_shifts, reduced_phis
    

    def enter_orientation_parameters(self, orientation_parameters, phi, psi, x_shift, y_shift, each_segment,
    inplane_angle, pixelsize, model_id=None):
        if model_id is None:
            model_id = each_segment.model_id
        fitted_param = orientation_parameters(each_segment.stack_id, each_segment.stack_id, model_id, phi,
        each_segment.theta, psi, x_shift / pixelsize, y_shift / pixelsize, inplane_angle, each_segment.peak,
        each_segment.mirror, None)
        
        return fitted_param
    

    def setup_parameter_containers(self):
        previous_params = []
        mask_params = []
        orientation_parameters = self.make_named_tuple_of_orientation_parameters()
        masking_parameters = self.make_named_tuple_of_masking_parameters()

        return orientation_parameters, previous_params, mask_params, masking_parameters


    def get_helically_related_parameters_for_first_refinement_cycle(self, temp_db, each_info,
    included_nonorientation_segments):
        orientation_parameters, previous_params, mask_params, masking_parameters = self.setup_parameter_containers()
        
        session = SpringDataBase().setup_sqlite_db(base, temp_db)
        orientation_results = session.query(SegmentTable).order_by(asc(SegmentTable.id)).all()
        session.close()
        os.remove(temp_db)

        segment_count = len(orientation_results)
#         helix_x_shifts = np.zeros(segment_count)
        helix_x_shifts = 2.0 * np.random.randn(segment_count)
        inplane_angles = np.array([each_segment.inplane_angle for each_segment in orientation_results])
        thetas = np.int64(90 + 4 * np.random.randn(segment_count))

        if len(self.helical_symmetries) > 1:
            azimuthal_increment = 360.0 / self.azimuthal_angle_count
            linear_phis = (np.arange(0, segment_count * azimuthal_increment, azimuthal_increment)) % 360
            phis = linear_phis[np.random.randint(0, segment_count, segment_count)]
            helix_y_shifts = 10.0 * np.random.randn(segment_count)
            model_ids = np.random.randint(0, len(self.helical_symmetries), segment_count)
        else:
            rise, rotation = self.helical_symmetries[0]
            distances_along_helix = self.stepsize * np.arange(segment_count)
            multiples_of_rise = np.int64(distances_along_helix / rise)
            phis = (multiples_of_rise * rotation) % 360
            helix_y_shifts = distances_along_helix - multiples_of_rise * rise
            model_ids = np.zeros(segment_count)

        sx, sy = SegClassReconstruct().compute_sx_sy_from_shifts_normal_and_parallel_to_helix_axis_with_inplane_angle(
        helix_x_shifts / each_info.pixelsize, helix_y_shifts / each_info.pixelsize, inplane_angles)

        for local_id, each_result in enumerate(orientation_results):
            if not self.unbending:
                psi = (-(each_result.inplane_angle + 270)) % 360
            elif self.unbending:
                psi = 270.0

            if each_result.stack_id in included_nonorientation_segments:
                ten_params = orientation_parameters(each_result.stack_id, local_id, int(model_ids[local_id]),
                phis[local_id], thetas[local_id], psi, sx[local_id], sy[local_id], each_result.inplane_angle, 0, 0, None)

                previous_params.append(ten_params)
                mask_params.append(masking_parameters(each_result.stack_id, local_id, each_result.inplane_angle, 0, 0))
        
        return previous_params, mask_params


    def get_zero_parameters_for_initial_stages_of_refinement(self, temp_db, included_nonorientation_segments):
        orientation_parameters, previous_params, mask_params, masking_parameters = self.setup_parameter_containers()
        
        session = SpringDataBase().setup_sqlite_db(base, temp_db)
        orientation_results = session.query(SegmentTable).order_by(asc(SegmentTable.id)).all()
        session.close()
        os.remove(temp_db)

        for local_id, each_result in enumerate(orientation_results):
            if not self.unbending:
                psi = (-(each_result.inplane_angle + 270)) % 360
            elif self.unbending:
                psi = 270.0

            if each_result.stack_id in included_nonorientation_segments:
                ten_params = orientation_parameters(each_result.stack_id, local_id, None, 0, 90, psi, 0, 0, 
                    each_result.inplane_angle, 0, 0, None)
                previous_params.append(ten_params)
                mask_params.append(masking_parameters(each_result.stack_id, local_id, each_result.inplane_angle, 0, 0))
        
        return previous_params, mask_params


    def get_refined_parameters_for_advanced_stages_of_refinement(self, each_info, reference_files,
    included_nonorientation_segments, temp_ref_db, last_cycle, ref_session):
        orientation_parameters, previous_params, mask_params, masking_parameters = self.setup_parameter_containers()

        if len(reference_files) > 1:
            model_id = None
        else:
            model_id = 0

        orientation_results = ref_session.query(RefinementCycleSegmentTable).\
        filter(RefinementCycleSegmentTable.cycle_id == last_cycle.id).\
        order_by(asc(RefinementCycleSegmentTable.stack_id)).all()

        ref_session.close()
        os.remove(temp_ref_db)

        for each_segment in orientation_results:
            if each_segment.stack_id in included_nonorientation_segments:
                if self.unbending:
                    ten_params = self.enter_orientation_parameters(orientation_parameters, 
                        each_segment.phi, 270.0, each_segment.unbent_shift_x_A, each_segment.unbent_shift_y_A, 
                        each_segment, each_segment.unbent_ip_angle, each_info.pixelsize, model_id)
                else:
                    ten_params = self.enter_orientation_parameters(orientation_parameters, 
                        each_segment.phi, each_segment.psi, each_segment.shift_x_A, each_segment.shift_y_A, 
                        each_segment, each_segment.inplane_angle, each_info.pixelsize, model_id)
                previous_params.append(ten_params)

                mask_params = self.get_masking_parameters(masking_parameters, mask_params, each_segment, 
                each_info.pixelsize)
        
        return previous_params, mask_params


    def prepare_previous_parameters_either_from_inplane_angle_or_from_previous_cycle(self, each_info, info_series,
    ref_cycle_id, each_iteration_number, reference_files):
        temp_db = self.copy_spring_db_to_tempdir()

        included_nonorientation_segments, excluded_non_orientation_counts = \
        SegmentSelect().filter_non_orientation_parameters_based_on_selection_criteria(self, temp_db,
        keep_helices_together=True)
        
        prev_ref_cycle_id = ref_cycle_id - 1
        refine_db = 'refinement{0:03}.db'.format(prev_ref_cycle_id)
        if self.resume_refinement_option and not os.path.exists(refine_db):
            refine_db = 'refinement{0:03}.db'.format(prev_ref_cycle_id)
            assert os.path.exists(refine_db)
        
        if os.path.exists(refine_db):
            temp_ref_db = self.copy_ref_db_to_tempdir(prev_ref_cycle_id)
            last_cycle, ref_session = self.get_last_cycle_from_refinement_database(temp_ref_db)
        else:
            last_cycle = None
            
#         if ref_cycle_id == 1 and not self.unbending and not self.reference_option:
#             previous_params, mask_params = self.get_helically_related_parameters_for_first_refinement_cycle(temp_db,
#             each_info, included_nonorientation_segments)
        if each_info.bin_factor == info_series[0].bin_factor and not self.resume_refinement_option:
            previous_params, mask_params = self.get_zero_parameters_for_initial_stages_of_refinement(temp_db,
            included_nonorientation_segments)
        elif last_cycle is not None:
            previous_params, mask_params = self.get_refined_parameters_for_advanced_stages_of_refinement(each_info,
            reference_files, included_nonorientation_segments, temp_ref_db,
            last_cycle, ref_session)
                
        temp_prev_ref_db = os.path.join(self.tempdir, 'refinement{0:03}.db'.format(prev_ref_cycle_id))
        if os.path.exists(temp_prev_ref_db):
            os.remove(temp_prev_ref_db)
            
        return previous_params, mask_params
    
    
    def find_maximum_cross_correlation_for_local_in_plane_angle(self, masked_segment_stack, step_x, each_info,
    reference_rings, projection_parameters, polar_interpolation_parameters, previous_params, alignment_size):
        coarse_determined_params = []
        orientation_parameters = self.make_named_tuple_of_orientation_parameters()
        coarse_align_loginfo = []
        for each_first_params in previous_params:
            
            selected_ref_rings, selected_prj_params, selected_log_info = \
            self.select_reference_rings_according_to_angular_restraints(reference_rings, each_info,
            projection_parameters, each_first_params)
            
            segment_params = \
            self.match_each_image_against_all_projections_with_delta_in_plane_restraint(masked_segment_stack,
            each_first_params, selected_ref_rings, selected_prj_params, each_info, step_x, self.delta_in_plane_rotation,
            polar_interpolation_parameters, alignment_size)
            
            segment_params.insert(0, each_first_params.stack_id)
            segment_params.insert(1, each_first_params.local_id)
            segment_params.append(each_first_params.rank_id)
            ref_orientation_param = orientation_parameters._make(segment_params)
            coarse_determined_params.append(ref_orientation_param)
            
            coarse_align_loginfo += [[each_first_params.stack_id, each_first_params.local_id, each_first_params.model_id,
            each_first_params.phi, each_first_params.theta, each_first_params.psi, each_first_params.shift_x,
            each_first_params.shift_y, each_first_params.inplane_angle, each_first_params.peak,
            each_first_params.mirror, None, None, 'previous']]
            
            coarse_align_loginfo += [[ref_orientation_param.stack_id, ref_orientation_param.local_id,
            ref_orientation_param.model_id, ref_orientation_param.phi, ref_orientation_param.theta,
            ref_orientation_param.psi, ref_orientation_param.shift_x, ref_orientation_param.shift_y,
            ref_orientation_param.inplane_angle, ref_orientation_param.peak, ref_orientation_param.mirror,
            selected_log_info[0], selected_log_info[1], 'current']]
        
        msg = tabulate(coarse_align_loginfo, list(orientation_parameters._fields)[:-1] + ['phi_range', 'theta_range', 'cycle'])
        self.log.tlog('The following alignment parameters were determined:\n{0}'.format(msg))
        
        return coarse_determined_params


    def get_restraint_tuple(self):
        restraint_tuple = namedtuple('restraints', 'azimuthal_restraint out_of_plane_restraint x_range y_range pixelsize')
        
        return restraint_tuple
    

    def refine_local_inplane_angle(self, masked_segment_stack, step_x, each_info, x_range, y_range,
    delta_in_plane_rotation, step_angle, reference_rings, projection_parameters, polar_interpolation_parameters,
    coarse_determined_params, fine_projection_stack, alignment_size):
        ref_orientation_params = []
        restraint_tuple = self.get_restraint_tuple()
        tilt_range = 1.2 * abs(self.out_of_plane_tilt_angle_range[-1] - self.out_of_plane_tilt_angle_range[0]) 
        
        restraint_info = restraint_tuple(1.2 * 360.0 / self.azimuthal_angle_count, tilt_range /
        self.out_of_plane_tilt_angle_count, x_range, y_range, each_info.pixelsize)
        
        orientation_parameters = self.make_named_tuple_of_orientation_parameters()
        ref_align_loginfo = []
        for each_previous_params in coarse_determined_params:
            selected_ref_rings, selected_prj_params, selected_log_info = \
            self.select_reference_rings_according_to_angular_restraints(reference_rings, restraint_info,
            projection_parameters, each_previous_params)
            
            if fine_projection_stack is not None:
                selected_ref_rings, polar_interpolation_parameters = \
                self.prepare_reference_rings_from_projections(fine_projection_stack, selected_prj_params,
                alignment_size, each_info.max_range)
            
            segment_params = \
            self.refine_each_image_against_projections_including_delta_inplane_restraint(masked_segment_stack,
            each_previous_params, selected_ref_rings, selected_prj_params, restraint_info, step_x,
            step_angle, delta_in_plane_rotation, polar_interpolation_parameters, alignment_size)
                
            segment_params.insert(0, each_previous_params.stack_id)
            segment_params.insert(1, each_previous_params.local_id)
            
            if hasattr(self, 'rank'):
                segment_params.append(self.rank)
            else:
                segment_params.append(None)
            
            ref_align_loginfo += [list(each_previous_params)[:-1] + [None, None, 'coarse']]
            ref_align_loginfo += [segment_params[:-1] + [selected_log_info[0], selected_log_info[1], 'fine']]
            
            segment_params = orientation_parameters._make(segment_params)
            ref_orientation_params.append(segment_params)
        
        msg = tabulate(ref_align_loginfo, list(each_previous_params._fields)[:-1] + ['phi_range', 'theta_range', 'align type'])
        self.log.tlog('The following refined parameters were determined for in-plane ' + \
        'angles:\n{0}'.format(msg))
        
        return ref_orientation_params
    

    def convert_coarse_into_fine_parameters(self, coarse_params):
        fine_params = []
        for each_param in coarse_params:
            params = list(each_param)[:-1]
            if hasattr(self, 'comm'):
                params.append(self.rank)
            elif not hasattr(self, 'comm'):
                params.append(None)
            fine_params.append(params)
        
        return fine_params
    
        
    def get_prj_params_named_tuple(self):
        prj_named_tuple = namedtuple('prj_parameters', 'id model_id phi theta psi shift_x shift_y')

        return prj_named_tuple


    def get_prj_stack_name_with_ending(self, projection_stack, ending):
        azimuthal_prj_stack = os.path.splitext(projection_stack)[0] + ending + \
        os.path.splitext(projection_stack)[-1]
        
        return azimuthal_prj_stack


    def copy_out_angular_series(self, projection_stack, azimuthal_prj_stack, azimuthal_angles):
        img = EMData()
        for each_azimuth_id, each_azimuth in enumerate(azimuthal_angles):
            img.read_image(projection_stack, each_azimuth.id)
            img.write_image(azimuthal_prj_stack, each_azimuth_id)
            
        return azimuthal_prj_stack


    def get_azimuthal_angles_from_prj_params(self, projection_parameters, model_id=None, unique_angle='theta'):
        """
        >>> from spring.segment3d.refine.sr3d_main import SegmentRefine3d
        >>> s = SegmentRefine3d()
        >>> proj_params = s.generate_Euler_angles_for_projection(4, [-8, 8], 3, 22.04) 
        >>> proj_params = [[0]+ each_p for each_p in proj_params]
        >>> proj_tuples = s.convert_list_of_projection_parameters_to_prj_tuples(proj_params)
        >>> s.get_azimuthal_angles_from_prj_params(proj_tuples) #doctest: +NORMALIZE_WHITESPACE
        [prj_parameters(id=4, model_id=0, phi=0.0, theta=90.0, psi=270.0, shift_x=0.0, shift_y=0.0), 
        prj_parameters(id=5, model_id=0, phi=27.549999999999997, theta=90.0, psi=270.0, shift_x=0.0, shift_y=0.0), 
        prj_parameters(id=6, model_id=0, phi=55.099999999999994, theta=90.0, psi=270.0, shift_x=0.0, shift_y=0.0), 
        prj_parameters(id=7, model_id=0, phi=82.65, theta=90.0, psi=270.0, shift_x=0.0, shift_y=0.0)]
        >>> s.get_azimuthal_angles_from_prj_params(proj_tuples, 0) #doctest: +NORMALIZE_WHITESPACE
        [prj_parameters(id=4, model_id=0, phi=0.0, theta=90.0, psi=270.0, shift_x=0.0, shift_y=0.0), 
        prj_parameters(id=5, model_id=0, phi=27.549999999999997, theta=90.0, psi=270.0, shift_x=0.0, shift_y=0.0), 
        prj_parameters(id=6, model_id=0, phi=55.099999999999994, theta=90.0, psi=270.0, shift_x=0.0, shift_y=0.0), 
        prj_parameters(id=7, model_id=0, phi=82.65, theta=90.0, psi=270.0, shift_x=0.0, shift_y=0.0)]
        >>> s.get_azimuthal_angles_from_prj_params(proj_tuples, 1)
        []
        """
        if model_id is None:
            un_thetas = np.unique([getattr(each_param, unique_angle) for each_param in projection_parameters])
            theta_closest_to_avg = np.argmin(np.abs(un_thetas - np.average(un_thetas)))

            azimuthal_angles = [each_param for each_param in projection_parameters \
                                if getattr(each_param, unique_angle) == un_thetas[theta_closest_to_avg]]
        else:
            un_thetas = np.unique([each_param.theta for each_param in projection_parameters
                                   if each_param.model_id == model_id])

            if len(un_thetas) > 0: 
                theta_closest_to_avg = np.argmin(np.abs(un_thetas - np.average(un_thetas)))
    
                azimuthal_angles = [each_param for each_param in projection_parameters \
                                    if getattr(each_param, unique_angle) == un_thetas[theta_closest_to_avg] and \
                                    each_param.model_id == model_id]
            else:
                azimuthal_angles = []

        return azimuthal_angles


    def get_out_of_plane_angles_from_prj_params(self, projection_parameters, model_id=None):
        """
        >>> from spring.segment3d.refine.sr3d_main import SegmentRefine3d
        >>> s = SegmentRefine3d()
        >>> proj_params = s.generate_Euler_angles_for_projection(4, [-8, 8], 3, 22.04) 
        >>> proj_params = [[0]+ each_p for each_p in proj_params]
        >>> proj_tuples = s.convert_list_of_projection_parameters_to_prj_tuples(proj_params)
        >>> s.get_out_of_plane_angles_from_prj_params(proj_tuples) #doctest: +NORMALIZE_WHITESPACE
        [prj_parameters(id=2, model_id=0, phi=55.099999999999994, theta=82.0, psi=270.0, shift_x=0.0, shift_y=0.0), 
        prj_parameters(id=6, model_id=0, phi=55.099999999999994, theta=90.0, psi=270.0, shift_x=0.0, shift_y=0.0), 
        prj_parameters(id=10, model_id=0, phi=55.099999999999994, theta=98.0, psi=270.0, shift_x=0.0, shift_y=0.0)]
        >>> s.get_out_of_plane_angles_from_prj_params(proj_tuples, 1)
        []
        """
        out_of_plane_angles = self.get_azimuthal_angles_from_prj_params(projection_parameters,model_id=model_id,
        unique_angle='phi')

        return out_of_plane_angles


    def remove_projection_stacks_and_copy_out_azimuthal_and_out_of_plane_series(self, prj_info):
        if hasattr(self, 'comm'):
            self.comm.barrier()
        if prj_info.projection_stack.startswith(self.tempdir):
            azimuthal_angles = self.get_azimuthal_angles_from_prj_params(prj_info.projection_parameters)
            out_of_plane_angles = self.get_out_of_plane_angles_from_prj_params(prj_info.projection_parameters)
            azimuthal_prj_stack = self.get_prj_stack_name_with_ending(prj_info.projection_stack, ending='az')
            out_of_plane_prj_stack = self.get_prj_stack_name_with_ending(prj_info.projection_stack, ending='out')

            if prj_info.fine_projection_stack is None:
                self.copy_out_angular_series(prj_info.projection_stack, azimuthal_prj_stack, azimuthal_angles)
                self.copy_out_angular_series(prj_info.projection_stack, out_of_plane_prj_stack, out_of_plane_angles)
            elif prj_info.fine_projection_stack is not None:
                self.copy_out_angular_series(prj_info.fine_projection_stack, azimuthal_prj_stack, azimuthal_angles)
                self.copy_out_angular_series(prj_info.fine_projection_stack, out_of_plane_prj_stack, out_of_plane_angles)
                os.remove(prj_info.fine_projection_stack)

            os.remove(prj_info.projection_stack)


    def convert_list_of_projection_parameters_to_prj_tuples(self, projection_parameters):
        prj_named_tuple = self.get_prj_params_named_tuple()
        
        prj_tuples = [ prj_named_tuple._make([each_prj_id] + each_prj_param) \
                      for each_prj_id, each_prj_param in enumerate(projection_parameters) ]

        return prj_tuples


    def perform_coarse_and_fine_projection_matching(self, each_info, masked_segment_stack, prj_info, previous_params,
    alignment_size):
        
        self.log.fcttolog()
        self.log.in_progress_log()

        prj_tuples = self.convert_list_of_projection_parameters_to_prj_tuples(prj_info.projection_parameters)
        prj_info = prj_info._replace(projection_parameters = prj_tuples)

        step_x = 1.0
        step_angle = 0.05
        zoom_refinement_factor = 5.0
        
        reference_rings, polar_interpolation_parameters = \
        self.prepare_reference_rings_from_projections(prj_info.projection_stack, prj_info.projection_parameters,
        alignment_size, each_info.max_range)
        
        coarse_determined_params = self.find_maximum_cross_correlation_for_local_in_plane_angle(masked_segment_stack,
        step_x, each_info, reference_rings, prj_info.projection_parameters, polar_interpolation_parameters,
        previous_params, alignment_size)
        
        fine_x_search_range = each_info.x_range / zoom_refinement_factor
        fine_y_search_range = each_info.y_range / zoom_refinement_factor
        fine_step_angle = np.sin(np.deg2rad(step_angle / zoom_refinement_factor))
        
        if prj_info.fine_projection_stack is not None:
            prj_tuples = self.convert_list_of_projection_parameters_to_prj_tuples(prj_info.fine_projection_parameters)
            prj_info = prj_info._replace(projection_parameters = prj_tuples)

            reference_rings = None
            polar_interpolation_parameters = None
        
        refined_orientation_parameters = self.refine_local_inplane_angle(masked_segment_stack,
        step_x / zoom_refinement_factor, each_info, fine_x_search_range, fine_y_search_range,
        self.delta_in_plane_rotation / zoom_refinement_factor, fine_step_angle, reference_rings, 
        prj_info.projection_parameters, polar_interpolation_parameters, coarse_determined_params, 
        prj_info.fine_projection_stack, alignment_size)
        
        self.remove_projection_stacks_and_copy_out_azimuthal_and_out_of_plane_series(prj_info)
            
        return refined_orientation_parameters, step_x, prj_info
    
